from __future__ import annotations
import contextlib
from typing import TYPE_CHECKING, Any, Callable, Iterator
import numpy as np

if TYPE_CHECKING:
    import torch
    from numpy.random import Generator
    from scipy.stats._qmc import QMCEngine
    from .config.candidate_rv import CandidateRV
from .config.raasp_driver import RAASPDriver


@contextlib.contextmanager
def record_duration(set_dt: Callable[[float], None]) -> Iterator[None]:
    import time

    t0 = time.perf_counter()
    try:
        yield
    finally:
        set_dt(time.perf_counter() - t0)


@contextlib.contextmanager
def torch_seed_context(
    seed: int, device: torch.device | Any | None = None
) -> Iterator[None]:
    import torch

    devices: list[int] | None = None
    if device is not None and getattr(device, "type", None) == "cuda":
        idx = 0 if getattr(device, "index", None) is None else int(device.index)
        devices = [idx]
    with torch.random.fork_rng(devices=devices, enabled=True):
        torch.manual_seed(int(seed))
        if device is not None and getattr(device, "type", None) == "cuda":
            torch.cuda.manual_seed_all(int(seed))
        if device is not None and getattr(device, "type", None) == "mps":
            if hasattr(torch, "mps") and hasattr(torch.mps, "manual_seed"):
                torch.mps.manual_seed(int(seed))
        yield


def get_gp_posterior_suppress_warning(model: Any, x_torch: Any) -> Any:
    import warnings

    try:
        from gpytorch.utils.warnings import GPInputWarning
    except Exception:
        GPInputWarning = None
    if GPInputWarning is None:
        return model.posterior(x_torch)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"The input matches the stored training data\..*",
            category=GPInputWarning,
        )
        return model.posterior(x_torch)


def latin_hypercube(
    num_points: int, num_dim: int, *, rng: Generator | Any
) -> np.ndarray:
    x = (1.0 + 2.0 * np.arange(0.0, num_points)) / float(2 * num_points)
    x = np.stack([x[rng.permutation(num_points)] for _ in range(num_dim)], axis=1)
    x += rng.uniform(-1.0, 1.0, size=(num_points, num_dim)) / float(2 * num_points)
    assert x.shape == (num_points, num_dim)
    return x


def argmax_random_tie(values: np.ndarray | Any, *, rng: Generator | Any) -> int:
    if values.ndim != 1:
        raise ValueError(values.shape)
    max_val = float(np.max(values))
    idx = np.nonzero(values >= max_val)[0]
    if idx.size == 0:
        return int(rng.integers(values.size))
    if idx.size == 1:
        return int(idx[0])
    j = int(rng.integers(idx.size))
    return int(idx[j])


def sobol_perturb_np(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    mask: np.ndarray | Any,
    *,
    sobol_engine: QMCEngine | Any,
) -> np.ndarray:
    n = num_candidates
    n_sobol = 1 if n <= 0 else 1 << (n - 1).bit_length()
    sobol_samples = sobol_engine.random(n_sobol)[:num_candidates]
    lb_array = np.asarray(lb)
    ub_array = np.asarray(ub)
    pert = lb_array + (ub_array - lb_array) * sobol_samples
    candidates = np.tile(x_center, (num_candidates, 1))
    if np.any(mask):
        candidates[mask] = pert[mask]
    return candidates


def uniform_perturb_np(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    mask: np.ndarray | Any,
    *,
    rng: Generator | Any,
) -> np.ndarray:
    lb_array = np.asarray(lb)
    ub_array = np.asarray(ub)
    pert = lb_array + (ub_array - lb_array) * rng.uniform(
        0.0, 1.0, size=(num_candidates, x_center.shape[-1])
    )
    candidates = np.tile(x_center, (num_candidates, 1))
    if np.any(mask):
        candidates[mask] = pert[mask]
    return candidates


def raasp_perturb(
    x_center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    num_pert: int = 20,
    rng: Generator | Any,
    candidate_rv: CandidateRV,
    sobol_engine: QMCEngine | Any | None = None,
) -> np.ndarray:
    num_dim = x_center.shape[-1]
    prob_perturb = min(num_pert / num_dim, 1.0)
    # Use binomial to determine how many dimensions to perturb for each candidate
    ks = np.maximum(rng.binomial(num_dim, prob_perturb, size=num_candidates), 1)
    mask = np.zeros((num_candidates, num_dim), dtype=bool)
    for i in range(num_candidates):
        idx = rng.choice(num_dim, size=ks[i], replace=False)
        mask[i, idx] = True

    from .config.candidate_rv import CandidateRV

    if candidate_rv == CandidateRV.SOBOL:
        if sobol_engine is None:
            raise ValueError("sobol_engine required for CandidateRV.SOBOL")
        return sobol_perturb_np(
            x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine
        )
    return uniform_perturb_np(x_center, lb, ub, num_candidates, mask, rng=rng)


def generate_raasp_candidates(
    center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    rng: Generator | Any,
    candidate_rv: CandidateRV,
    sobol_engine: QMCEngine | Any | None = None,
    num_pert: int = 20,
) -> np.ndarray:
    if num_candidates <= 0:
        raise ValueError(num_candidates)
    return raasp_perturb(
        center,
        lb,
        ub,
        num_candidates,
        num_pert=num_pert,
        rng=rng,
        candidate_rv=candidate_rv,
        sobol_engine=sobol_engine,
    )


def generate_raasp_candidates_uniform(
    center: np.ndarray | Any,
    lb: np.ndarray | list[float] | Any,
    ub: np.ndarray | list[float] | Any,
    num_candidates: int,
    *,
    rng: Generator | Any,
    num_pert: int = 20,
) -> np.ndarray:
    from .config.candidate_rv import CandidateRV

    return generate_raasp_candidates(
        center,
        lb,
        ub,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.UNIFORM,
        num_pert=num_pert,
    )


def to_unit(x: np.ndarray | Any, bounds: np.ndarray | Any) -> np.ndarray:
    lb = bounds[:, 0]
    ub = bounds[:, 1]
    if np.any(ub <= lb):
        raise ValueError(bounds)
    return (x - lb) / (ub - lb)


def from_unit(x_unit: np.ndarray | Any, bounds: np.ndarray | Any) -> np.ndarray:
    lb = np.asarray(bounds[:, 0])
    ub = np.asarray(bounds[:, 1])
    return lb + x_unit * (ub - lb)


def gp_thompson_sample(
    model: Any,
    x_cand: np.ndarray | Any,
    num_arms: int,
    rng: Generator | Any,
    *,
    gp_y_mean: float,
    gp_y_std: float,
) -> np.ndarray:
    import gpytorch
    import torch

    x_torch = torch.as_tensor(x_cand, dtype=torch.float64)
    seed = int(rng.integers(2**31 - 1))
    with (
        torch.no_grad(),
        gpytorch.settings.fast_pred_var(),
        torch_seed_context(seed, device=x_torch.device),
    ):
        posterior = model.posterior(x_torch)
        samples = posterior.sample(sample_shape=torch.Size([1]))
    if samples.ndim != 2:
        raise ValueError(samples.shape)
    ts = samples[0].reshape(-1)
    scores = ts.detach().cpu().numpy().reshape(-1)
    scores = gp_y_mean + gp_y_std * scores
    shuffled_indices = rng.permutation(len(scores))
    shuffled_scores = scores[shuffled_indices]
    top_k_in_shuffled = np.argpartition(-shuffled_scores, num_arms - 1)[:num_arms]
    idx = shuffled_indices[top_k_in_shuffled]
    return idx


def compute_full_box_bounds_1d(
    x_center: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    lb = np.zeros_like(x_center, dtype=float)
    ub = np.ones_like(x_center, dtype=float)
    return lb, ub


def get_single_incumbent_index(
    selector: Any,
    y: np.ndarray,
    rng: Generator,
    mu: np.ndarray | None = None,
) -> np.ndarray:
    y = np.asarray(y, dtype=float)
    if y.size == 0:
        return np.array([], dtype=int)
    best_idx = selector.select(y, mu, rng)
    return np.array([best_idx])


def get_incumbent_index(
    selector: Any,
    y: np.ndarray,
    rng: Generator,
    mu: np.ndarray | None = None,
) -> int:
    y = np.asarray(y, dtype=float)
    if y.size == 0:
        raise ValueError("y is empty")
    return int(selector.select(y, mu, rng))


def get_scalar_incumbent_value(
    selector: Any,
    y_obs: np.ndarray,
    rng: Generator,
    *,
    mu_obs: np.ndarray | None = None,
) -> np.ndarray:
    y = np.asarray(y_obs, dtype=float)
    if y.size == 0:
        return np.array([], dtype=float)
    idx = get_incumbent_index(selector, y, rng, mu=mu_obs)
    use_mu = bool(getattr(selector, "noise_aware", False))
    values = mu_obs if use_mu else y
    if values is None:
        raise ValueError("noise_aware incumbent selection requires mu_obs")
    v = np.asarray(values, dtype=float)
    if v.ndim == 2:
        value = float(v[idx, 0])
    elif v.ndim == 1:
        value = float(v[idx])
    else:
        raise ValueError(v.shape)
    return np.array([value], dtype=float)


class ScalarIncumbentMixin:
    incumbent_selector: Any

    def get_incumbent_index(
        self,
        y: np.ndarray | Any,
        rng: Generator,
        mu: np.ndarray | None = None,
    ) -> int:
        return get_incumbent_index(self.incumbent_selector, y, rng, mu=mu)

    def get_incumbent_value(
        self,
        y_obs: np.ndarray | Any,
        rng: Generator,
        mu_obs: np.ndarray | None = None,
    ) -> np.ndarray:
        return get_scalar_incumbent_value(
            self.incumbent_selector, y_obs, rng, mu_obs=mu_obs
        )


def generate_tr_candidates_orig(
    compute_bounds_1d: Any,
    x_center: np.ndarray,
    lengthscales: np.ndarray | None,
    num_candidates: int,
    *,
    rng: Generator,
    candidate_rv: CandidateRV,
    sobol_engine: QMCEngine | None = None,
    num_pert: int = 20,
) -> np.ndarray:
    from .config.candidate_rv import CandidateRV

    lb, ub = compute_bounds_1d(x_center, lengthscales)
    if candidate_rv == CandidateRV.SOBOL:
        if sobol_engine is None:
            raise ValueError(
                "sobol_engine is required when candidate_rv=CandidateRV.SOBOL"
            )
        return generate_raasp_candidates(
            x_center,
            lb,
            ub,
            num_candidates,
            rng=rng,
            candidate_rv=candidate_rv,
            sobol_engine=sobol_engine,
            num_pert=num_pert,
        )
    if candidate_rv == CandidateRV.UNIFORM:
        return generate_raasp_candidates_uniform(
            x_center, lb, ub, num_candidates, rng=rng, num_pert=num_pert
        )
    raise ValueError(candidate_rv)


def generate_tr_candidates_fast(
    compute_bounds_1d: Any,
    x_center: np.ndarray,
    lengthscales: np.ndarray | None,
    num_candidates: int,
    *,
    rng: Generator,
    candidate_rv: CandidateRV,
    num_pert: int,
) -> np.ndarray:
    from scipy.stats import qmc
    from .config.candidate_rv import CandidateRV

    lb, ub = compute_bounds_1d(x_center, lengthscales)
    num_dim = x_center.shape[-1]
    candidates = np.tile(x_center, (num_candidates, 1))
    prob_perturb = min(num_pert / num_dim, 1.0)
    ks = np.maximum(rng.binomial(num_dim, prob_perturb, size=num_candidates), 1)
    max_k = int(np.max(ks))
    samples = (
        qmc.Sobol(d=max_k, scramble=True, seed=int(rng.integers(0, 2**31))).random(
            num_candidates
        )
        if candidate_rv == CandidateRV.SOBOL
        else rng.random((num_candidates, max_k))
    )
    for i in range(num_candidates):
        idx = rng.choice(num_dim, size=ks[i], replace=False)
        candidates[i, idx] = lb[idx] + (ub[idx] - lb[idx]) * samples[i, : ks[i]]
    assert candidates.shape == (num_candidates, num_dim)
    return candidates


def generate_tr_candidates(
    compute_bounds_1d: Any,
    x_center: np.ndarray,
    lengthscales: np.ndarray | None,
    num_candidates: int,
    *,
    rng: Generator,
    candidate_rv: CandidateRV,
    sobol_engine: QMCEngine | None,
    raasp_driver: RAASPDriver,
    num_pert: int,
) -> np.ndarray:
    from .config.raasp_driver import RAASPDriver

    if raasp_driver == RAASPDriver.FAST:
        return generate_tr_candidates_fast(
            compute_bounds_1d,
            x_center,
            lengthscales,
            num_candidates,
            rng=rng,
            candidate_rv=candidate_rv,
            num_pert=num_pert,
        )
    return generate_tr_candidates_orig(
        compute_bounds_1d,
        x_center,
        lengthscales,
        num_candidates,
        rng=rng,
        candidate_rv=candidate_rv,
        sobol_engine=sobol_engine,
        num_pert=num_pert,
    )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from ..optimizer import Optimizer
    from ..types import TellInputs


class OptimizationStrategy(ABC):
    @abstractmethod
    def ask(self, opt: Optimizer, num_arms: int) -> np.ndarray: ...
    @abstractmethod
    def tell(
        self, opt: Optimizer, inputs: TellInputs, *, x_unit: np.ndarray
    ) -> np.ndarray: ...
    @abstractmethod
    def init_progress(self) -> tuple[int, int] | None: ...



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    pass


@dataclass(frozen=True)
class MultiObjectiveConfig:
    num_metrics: int
    alpha: float = 0.05

    def __post_init__(self) -> None:
        if self.num_metrics < 2:
            raise ValueError(
                f"num_metrics must be >= 2 for MORBO, got {self.num_metrics}"
            )
        if self.alpha <= 0:
            raise ValueError(f"alpha must be > 0, got {self.alpha}")



from __future__ import annotations
import numpy as np
import pytest
from enn.enn.enn_index import ENNIndex
from enn.enn.enn_hash import (
    normal_hash_batch_multi_seed,
    normal_hash_batch_multi_seed_fast,
)


def test_enn_index_init_and_search():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    x_scale = np.ones((1, 3), dtype=float)
    index = ENNIndex(train_x, num_dim=3, x_scale=x_scale, scale_x=False)
    query = rng.standard_normal((5, 3))
    dist2s, idx = index.search(query, search_k=3, exclude_nearest=False)
    assert dist2s.shape == (5, 3) and idx.shape == (5, 3)
    assert np.all(idx >= 0) and np.all(idx < 20)


def test_enn_index_search_exclude_nearest():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    x_scale = np.ones((1, 3), dtype=float)
    index = ENNIndex(train_x, num_dim=3, x_scale=x_scale, scale_x=False)
    query = train_x[:3]
    dist2s_include, idx_include = index.search(query, search_k=3, exclude_nearest=False)
    dist2s_exclude, idx_exclude = index.search(query, search_k=3, exclude_nearest=True)
    assert dist2s_include.shape == (3, 3) and dist2s_exclude.shape == (3, 2)
    assert np.allclose(dist2s_include[:, 0], 0.0, atol=1e-6)


def test_enn_index_with_scaling():
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    x_scale = np.array([[2.0, 0.5, 1.0]])
    index = ENNIndex(train_x / x_scale, num_dim=3, x_scale=x_scale, scale_x=True)
    query = rng.standard_normal((5, 3))
    dist2s, idx = index.search(query, search_k=3, exclude_nearest=False)
    assert dist2s.shape == (5, 3) and idx.shape == (5, 3)


@pytest.mark.parametrize("query_shape,search_k", [((5, 3), 0), ((5, 4), 3)])
def test_enn_index_search_invalid_inputs(query_shape, search_k):
    rng = np.random.default_rng(42)
    train_x = rng.standard_normal((20, 3))
    index = ENNIndex(train_x, num_dim=3, x_scale=np.ones((1, 3)), scale_x=False)
    with pytest.raises(ValueError):
        index.search(
            rng.standard_normal(query_shape), search_k=search_k, exclude_nearest=False
        )


def test_normal_hash_batch_multi_seed_shape():
    function_seeds = np.array([1, 2, 3], dtype=np.int64)
    data_indices = np.array([[0, 1, 2], [3, 4, 5]], dtype=int)
    result = normal_hash_batch_multi_seed(function_seeds, data_indices, num_metrics=2)
    assert result.shape == (3, 2, 3, 2)


def test_normal_hash_batch_multi_seed_deterministic():
    function_seeds = np.array([42], dtype=np.int64)
    data_indices = np.array([[0, 1]], dtype=int)
    result1 = normal_hash_batch_multi_seed(function_seeds, data_indices, num_metrics=1)
    result2 = normal_hash_batch_multi_seed(function_seeds, data_indices, num_metrics=1)
    assert np.allclose(result1, result2)


def test_normal_hash_batch_multi_seed_different_seeds():
    data_indices = np.array([[0, 1]], dtype=int)
    result1 = normal_hash_batch_multi_seed(
        np.array([1], dtype=np.int64), data_indices, num_metrics=1
    )
    result2 = normal_hash_batch_multi_seed(
        np.array([2], dtype=np.int64), data_indices, num_metrics=1
    )
    assert not np.allclose(result1, result2)


def test_normal_hash_batch_multi_seed_fast_shape_and_deterministic():
    function_seeds = np.array([1, 2, 3], dtype=np.int64)
    data_indices = np.array([[0, 1, 2], [3, 4, 5]], dtype=int)
    out1 = normal_hash_batch_multi_seed_fast(
        function_seeds, data_indices, num_metrics=2
    )
    out2 = normal_hash_batch_multi_seed_fast(
        function_seeds, data_indices, num_metrics=2
    )
    assert out1.shape == (3, 2, 3, 2)
    assert np.allclose(out1, out2)


def test_normal_hash_batch_multi_seed_fast_different_seeds():
    data_indices = np.array([[0, 1]], dtype=int)
    out1 = normal_hash_batch_multi_seed_fast(
        np.array([1], dtype=np.int64), data_indices, num_metrics=3
    )
    out2 = normal_hash_batch_multi_seed_fast(
        np.array([2], dtype=np.int64), data_indices, num_metrics=3
    )
    assert not np.allclose(out1, out2)



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING
from .enn_index_driver import ENNIndexDriver
from .enn_fit_config import ENNFitConfig

if TYPE_CHECKING:
    from ..components.protocols import Surrogate


@dataclass(frozen=True)
class ENNSurrogateConfig:
    k: int | None = None
    fit: ENNFitConfig = ENNFitConfig()
    scale_x: bool = False
    index_driver: ENNIndexDriver = ENNIndexDriver.FLAT

    @property
    def num_fit_samples(self) -> int | None:
        return self.fit.num_fit_samples

    @property
    def num_fit_candidates(self) -> int | None:
        return self.fit.num_fit_candidates

    def build(self) -> Surrogate:
        from ..components.builder import build_surrogate

        return build_surrogate(self)



from __future__ import annotations

import numpy as np
import pytest
from scipy.stats import qmc

from enn.turbo.config.morbo_tr_config import (
    MorboTRConfig,
    MultiObjectiveConfig,
    RescalePolicyConfig,
)
from enn.turbo.config.no_tr_config import NoTRConfig
from enn.turbo.config import Rescalarize
from enn.turbo.config.turbo_tr_config import TRLengthConfig, TurboTRConfig
from enn.turbo.morbo_trust_region import MorboTrustRegion
from enn.turbo.no_trust_region import NoTrustRegion
from enn.turbo.turbo_utils import (
    compute_full_box_bounds_1d,
)
from enn.turbo.turbo_trust_region import TurboTrustRegion


def test_no_trust_region_init():
    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    assert tr.num_dim == 3 and tr.length == 1.0


def test_no_trust_region_update_does_nothing():
    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    tr.update(np.array([1.0, 2.0, 3.0]), np.array([3.0]))
    assert tr.length == 1.0


def test_no_trust_region_needs_restart():
    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    assert not tr.needs_restart()


def test_no_trust_region_validate_request():
    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    tr.validate_request(4)
    tr.validate_request(5)


def test_no_trust_region_compute_bounds_1d():
    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    x_center = np.array([0.5, 0.5, 0.5])
    lb, ub = tr.compute_bounds_1d(x_center)
    assert np.allclose(lb, 0.0) and np.allclose(ub, 1.0)


def test_no_trust_region_generate_candidates():
    from enn.turbo.config import CandidateRV, RAASPDriver
    from enn.turbo.turbo_utils import generate_tr_candidates

    config = NoTRConfig()
    tr = NoTrustRegion(config=config, num_dim=3)
    rng = np.random.default_rng(42)
    sobol = qmc.Sobol(d=3, scramble=True, seed=42)
    x_center = np.array([0.5, 0.5, 0.5])
    candidates = generate_tr_candidates(
        tr.compute_bounds_1d,
        x_center,
        None,
        100,
        rng=rng,
        candidate_rv=CandidateRV.SOBOL,
        sobol_engine=sobol,
        raasp_driver=RAASPDriver.ORIG,
        num_pert=20,
    )
    assert candidates.shape == (100, 3)
    assert np.all(candidates >= 0.0) and np.all(candidates <= 1.0)


def test_turbo_trust_region_validate_request():
    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.8, length_min=0.5**7, length_max=1.6)
    )
    tr = TurboTrustRegion(config=config, num_dim=3)
    tr.validate_request(4)
    with pytest.raises(ValueError):
        tr.validate_request(5)


def test_turbo_trust_region_get_incumbent_indices():
    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.8, length_min=0.5**7, length_max=1.6)
    )
    tr = TurboTrustRegion(config=config, num_dim=3)
    rng = np.random.default_rng(42)
    y = np.array([1.0, 5.0, 3.0, 2.0, 4.0])
    indices = tr.get_incumbent_indices(y, rng)
    assert 1 in indices


def test_morbo_trust_region_validate_request():
    rng = np.random.default_rng(42)
    config = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        rescale_policy=RescalePolicyConfig(rescalarize=Rescalarize.ON_PROPOSE),
    )
    tr = MorboTrustRegion(config=config, num_dim=3, rng=rng)
    tr.validate_request(4)
    with pytest.raises(ValueError):
        tr.validate_request(5)


def test_morbo_trust_region_get_incumbent_indices():
    rng = np.random.default_rng(42)
    config = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        rescale_policy=RescalePolicyConfig(rescalarize=Rescalarize.ON_PROPOSE),
    )
    tr = MorboTrustRegion(config=config, num_dim=3, rng=rng)
    y = np.array([[1.0, 5.0], [5.0, 1.0], [3.0, 3.0], [2.0, 2.0]])
    indices = tr.get_incumbent_indices(y, rng)
    assert len(indices) >= 1


def test_compute_full_box_bounds_1d():
    lb, ub = compute_full_box_bounds_1d(np.array([0.25, 0.5, 0.75]))
    assert np.allclose(lb, 0.0) and np.allclose(ub, 1.0)


def test_turbo_trust_region_compute_bounds_1d_with_lengthscales():
    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.8, length_min=0.5**7, length_max=1.6)
    )
    tr = TurboTrustRegion(config=config, num_dim=3)
    x_center = np.array([0.5, 0.5, 0.5])
    lengthscales = np.array([0.5, 1.0, 2.0])
    lb, ub = tr.compute_bounds_1d(x_center, lengthscales=lengthscales)
    assert np.allclose(lb, [0.3, 0.1, 0.0])
    assert np.allclose(ub, [0.7, 0.9, 1.0])


def test_turbo_trust_region_expansion_and_contraction():
    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.4, length_min=0.1, length_max=1.6)
    )
    tr = TurboTrustRegion(config=config, num_dim=4)
    tr.validate_request(num_arms=4)
    assert tr.length == 0.4
    values = []
    for v in [1.0, 2.0, 3.0, 4.0]:
        values.append(v)
        y_obs = np.array(values, dtype=float)
        tr.update(y_obs, np.array([float(np.max(y_obs))], dtype=float))
    assert np.isclose(tr.length, 0.8), f"Expected 0.8 after expansion, got {tr.length}"
    for _ in range(tr.failure_tolerance):
        values.append(values[-1])
        y_obs = np.array(values, dtype=float)
        tr.update(y_obs, np.array([float(np.max(y_obs))], dtype=float))
    assert np.isclose(tr.length, 0.4), (
        f"Expected 0.4 after contraction, got {tr.length}"
    )



from __future__ import annotations
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING
import numpy as np
from ..sampling import draw_lhd
from ..types.appendable_array import AppendableArray
from .optimization_strategy import OptimizationStrategy

if TYPE_CHECKING:
    from ..optimizer import Optimizer
    from ..types import TellInputs


@dataclass
class TurboHybridStrategy(OptimizationStrategy):
    _bounds: np.ndarray
    _num_dim: int
    _rng: object
    _num_init: int
    _init_lhd: np.ndarray
    _init_idx: int = 0

    @classmethod
    def create(
        cls, *, bounds: np.ndarray, rng: object, num_init: int | None
    ) -> TurboHybridStrategy:
        from numpy.random import Generator

        bounds = np.asarray(bounds, dtype=float)
        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(f"bounds must be (d, 2), got {bounds.shape}")
        num_dim = int(bounds.shape[0])
        if not isinstance(rng, Generator):
            raise TypeError("rng must be a numpy.random.Generator")
        n_init = int(num_init if num_init is not None else 2 * num_dim)
        if n_init <= 0:
            raise ValueError(f"num_init must be > 0, got {n_init}")
        init_lhd = draw_lhd(bounds=bounds, num_arms=n_init, rng=rng)
        return cls(
            _bounds=bounds,
            _num_dim=num_dim,
            _rng=rng,
            _num_init=n_init,
            _init_lhd=init_lhd,
        )

    def _reset_init(self) -> None:
        from numpy.random import Generator

        if not isinstance(self._rng, Generator):
            raise TypeError("rng must be a numpy.random.Generator")
        self._init_lhd = draw_lhd(
            bounds=self._bounds, num_arms=self._num_init, rng=self._rng
        )
        self._init_idx = 0

    def _get_init_points(self, num_arms: int, *, fallback_fn=None) -> np.ndarray:
        num_arms = int(num_arms)
        num_to_return = min(num_arms, self._num_init - self._init_idx)
        result = self._init_lhd[self._init_idx : self._init_idx + num_to_return]
        self._init_idx += num_to_return
        if num_to_return < num_arms:
            extra = (
                fallback_fn
                or (lambda n: draw_lhd(bounds=self._bounds, num_arms=n, rng=self._rng))
            )(num_arms - num_to_return)
            result = np.vstack([result, extra])
        return result

    def ask(self, opt: Optimizer, num_arms: int) -> np.ndarray:
        if opt._tr_state.needs_restart():
            opt._tr_state.restart(opt._rng)
            opt._restart_generation += 1
            opt._x_obs = AppendableArray()
            opt._y_obs = AppendableArray()
            opt._yvar_obs = AppendableArray()
            opt._y_tr_list = []
            opt._incumbent_idx = None
            opt._incumbent_x_unit = None
            opt._incumbent_y_scalar = None
            self._reset_init()
            return self._get_init_points(num_arms)
        if self._init_idx < self._num_init:

            def fallback(n: int) -> np.ndarray:
                return opt._ask_normal(n, is_fallback=True)

            return self._get_init_points(
                num_arms,
                fallback_fn=fallback if len(opt._x_obs) > 0 else None,
            )
        if len(opt._x_obs) == 0:
            return draw_lhd(bounds=self._bounds, num_arms=num_arms, rng=self._rng)
        opt._tr_state.validate_request(int(num_arms))
        return opt._ask_normal(int(num_arms))

    def init_progress(self) -> tuple[int, int] | None:
        return (int(self._init_idx), int(self._num_init))

    def tell(
        self, opt: Optimizer, inputs: TellInputs, *, x_unit: np.ndarray
    ) -> np.ndarray:
        x_all = opt._x_obs.view()
        y_all = opt._y_obs.view()
        y_var_all = opt._yvar_obs.view() if len(opt._yvar_obs) > 0 else None
        t0 = time.perf_counter()
        opt._surrogate.fit(
            x_all, y_all, y_var_all, num_steps=opt._gp_num_steps, rng=opt._rng
        )
        opt._dt_fit = time.perf_counter() - t0
        opt._y_tr_list = y_all.tolist()
        opt._update_incumbent()
        try:
            new_posterior = opt._surrogate.predict(x_unit)
            y_estimate = np.asarray(new_posterior.mu, dtype=float)
        except RuntimeError:
            y_estimate = np.asarray(inputs.y, dtype=float)
        y_incumbent = opt._incumbent_y_scalar
        y_obs = y_all
        opt._tr_state.update(y_obs, y_incumbent)
        if opt._trailing_obs is not None:
            opt._trim_trailing_obs()
        return y_estimate



from __future__ import annotations
import conftest
import numpy as np
import pytest
from enn import create_optimizer
from enn.turbo.components import NoSurrogate, SurrogateResult
from enn.turbo.config import CandidateRV
from enn.turbo.optimizer import Optimizer
from enn.turbo.optimizer_config import turbo_enn_config, turbo_zero_config


def test_turbo_enn_affine_invariance_under_dynamic_y_range() -> None:
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    num_arms, num_steps = 4, 20
    config = turbo_enn_config()

    def run(global_scale: float, global_shift: float) -> tuple[np.ndarray, np.ndarray]:
        rng = np.random.default_rng(0)
        opt = create_optimizer(bounds=bounds, config=config, rng=rng)
        unit_trajectory: list[np.ndarray] = []
        tr_lengths: list[float] = []
        for t in range(num_steps):
            x = opt.ask(num_arms=num_arms)
            u = x.copy()
            base_y = conftest.sphere_objective(2.0 * u - 1.0)
            a_t = 10.0 ** float(np.sin(t + 1.0)) + 1e-6
            b_t = 1e6 * float(np.cos(t + 1.0))
            y = global_scale * (a_t * base_y + b_t) + global_shift
            opt.tell(x, y)
            unit_trajectory.append(u)
            tr_lengths.append(opt.tr_length)
        return np.stack(unit_trajectory, axis=0), np.asarray(tr_lengths, dtype=float)

    traj_a, lengths_a = run(global_scale=1.0, global_shift=0.0)
    traj_b, lengths_b = run(global_scale=3.7, global_shift=-0.2)
    assert np.allclose(traj_a, traj_b)
    assert np.allclose(lengths_a, lengths_b)


def test_trailing_obs_preserves_unique_best_and_is_deterministic_under_ties() -> None:
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    config = turbo_zero_config(
        trailing_obs=5, num_init=2, candidate_rv=CandidateRV.UNIFORM
    )
    num_steps = 30

    def run(y_fn) -> list[tuple[np.ndarray, np.ndarray]]:
        rng = np.random.default_rng(0)
        opt = create_optimizer(bounds=bounds, config=config, rng=rng)
        snapshots: list[tuple[np.ndarray, np.ndarray]] = []
        for t in range(num_steps):
            x = opt.ask(num_arms=1)
            y = np.asarray([float(y_fn(t))], dtype=float)
            opt.tell(x, y)
            snapshots.append(
                (
                    opt._x_obs.view().copy(),
                    opt._y_obs.view().copy(),
                )
            )
        return snapshots

    snaps_unique_best = run(lambda t: 1_000.0 if t == 0 else -float(t))
    for _, y_obs in snaps_unique_best[10:]:
        assert float(np.max(y_obs)) == 1_000.0
    snaps_ties_1 = run(lambda t: 0.0)
    snaps_ties_2 = run(lambda t: 0.0)
    assert len(snaps_ties_1) == len(snaps_ties_2)
    for (x1, y1), (x2, y2) in zip(snaps_ties_1, snaps_ties_2, strict=True):
        assert np.array_equal(y1, y2)
        assert np.allclose(x1, x2)


class _NoSurrogateWithLengthscales(NoSurrogate):
    def __init__(self, lengthscales: np.ndarray) -> None:
        super().__init__()
        self._lengthscales = np.asarray(lengthscales, dtype=float)

    def fit(
        self,
        x_obs: np.ndarray,
        y_obs: np.ndarray,
        y_var: np.ndarray | None = None,
        *,
        num_steps: int = 0,
        rng: object | None = None,
    ) -> SurrogateResult:
        super().fit(
            x_obs,
            y_obs,
            y_var,
            num_steps=num_steps,
            rng=rng,
        )
        return SurrogateResult(model=None, lengthscales=self._lengthscales)


class _FirstNAcqOptimizer:
    def select(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        surrogate: object,
        rng: object,
        *,
        tr_state: object | None = None,
    ) -> np.ndarray:
        return np.asarray(x_cand, dtype=float)[: int(num_arms)]


def test_candidate_generation_with_extreme_lengthscales_stays_in_bounds() -> None:
    bounds = np.array([[0.0, 1.0]] * 5, dtype=float)
    rng = np.random.default_rng(0)
    config = turbo_zero_config(
        num_init=1,
        num_candidates=64,
        candidate_rv=CandidateRV.UNIFORM,
    )
    surrogate = _NoSurrogateWithLengthscales(
        lengthscales=np.array([1e-6, 1e6, 1e-3, 1e3, 1.0], dtype=float)
    )
    opt = Optimizer(
        bounds=bounds,
        config=config,
        rng=rng,
        surrogate=surrogate,
        acquisition_optimizer=_FirstNAcqOptimizer(),
    )
    x0 = opt.ask(num_arms=1)
    opt.tell(x0, np.array([0.0], dtype=float))
    x_obs = opt._x_obs.view()
    y_obs = opt._y_obs.view()
    x_center = opt._find_x_center(x_obs, y_obs)
    assert x_center is not None
    x = opt.ask(num_arms=4)
    assert x.shape == (4, bounds.shape[0])
    assert np.all(np.isfinite(x))
    assert np.all(x >= 0.0) and np.all(x <= 1.0)
    assert np.all(np.any(np.abs(x - x_center.reshape(1, -1)) > 0.0, axis=1))


def test_candidate_generation_raises_on_nonfinite_lengthscales() -> None:
    bounds = np.array([[0.0, 1.0]] * 3, dtype=float)
    rng = np.random.default_rng(0)
    config = turbo_zero_config(
        num_init=1,
        num_candidates=16,
        candidate_rv=CandidateRV.UNIFORM,
    )
    surrogate = _NoSurrogateWithLengthscales(
        lengthscales=np.array([np.nan, 1.0, 1.0], dtype=float)
    )
    opt = Optimizer(
        bounds=bounds,
        config=config,
        rng=rng,
        surrogate=surrogate,
        acquisition_optimizer=_FirstNAcqOptimizer(),
    )
    x0 = opt.ask(num_arms=1)
    opt.tell(x0, np.array([0.0], dtype=float))
    with pytest.raises(ValueError, match="lengthscales"):
        _ = opt.ask(num_arms=2)



from __future__ import annotations
import numpy as np
from enn.turbo.config import (
    MorboTRConfig,
    MultiObjectiveConfig,
    NoTRConfig,
    TurboTRConfig,
    turbo_zero_config,
)
from enn.turbo.config.turbo_tr_config import TRLengthConfig
from enn.turbo.impl_helpers import (
    estimate_y_passthrough,
    get_x_center_fallback,
    handle_restart_check_multi_objective,
    handle_restart_clear_always,
)


def _build_and_verify_tr(tr_config, rng, num_dim=3):
    tr = tr_config.build(num_dim=num_dim, rng=rng)
    assert tr is not None
    assert hasattr(tr, "length")
    return tr


def test_trust_region_config_build_no_tr():
    _build_and_verify_tr(NoTRConfig(), np.random.default_rng(42))


def test_trust_region_config_build_turbo_tr():
    tr = _build_and_verify_tr(TurboTRConfig(), np.random.default_rng(42))
    assert tr.length == 0.8


def test_trust_region_config_build_morbo_tr():
    config = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2))
    _build_and_verify_tr(config, np.random.default_rng(42))


def test_get_x_center_fallback_empty():
    config = turbo_zero_config()
    rng = np.random.default_rng(42)
    result = get_x_center_fallback(config, [], [], rng)
    assert result is None


def test_get_x_center_fallback_single_objective():
    config = turbo_zero_config()
    rng = np.random.default_rng(42)
    x_obs = [[0.5, 0.5], [0.3, 0.7], [0.8, 0.2]]
    y_obs = [1.0, 3.0, 2.0]
    result = get_x_center_fallback(config, x_obs, y_obs, rng)
    assert result is not None
    assert np.allclose(result, [0.3, 0.7])


def test_get_x_center_fallback_with_tr_state():
    from enn.turbo.turbo_trust_region import TurboTrustRegion

    rng = np.random.default_rng(42)
    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.8, length_min=0.5**7, length_max=1.6)
    )
    tr_state = TurboTrustRegion(config=config, num_dim=2)
    x_obs = [[0.5, 0.5], [0.3, 0.7], [0.8, 0.2]]
    y_obs = [1.0, 3.0, 2.0]
    result = get_x_center_fallback(None, x_obs, y_obs, rng, tr_state=tr_state)
    assert result is not None
    assert np.allclose(result, [0.3, 0.7])


def test_handle_restart_clear_always():
    x = [1, 2, 3]
    y = [4, 5, 6]
    yvar = [0.1, 0.2, 0.3]
    cleared, idx = handle_restart_clear_always(x, y, yvar)
    assert cleared is True
    assert idx == 0
    assert x == []
    assert y == []
    assert yvar == []


def test_handle_restart_check_multi_objective_single():
    from enn.turbo.turbo_trust_region import TurboTrustRegion

    config = TurboTRConfig(
        length=TRLengthConfig(length_init=0.8, length_min=0.5**7, length_max=1.6)
    )
    tr_state = TurboTrustRegion(config=config, num_dim=2)
    x = [1, 2, 3]
    y = [4, 5, 6]
    yvar = [0.1, 0.2, 0.3]
    cleared, idx = handle_restart_check_multi_objective(
        tr_state, x, y, yvar, init_idx=5
    )
    assert cleared is False
    assert idx == 5
    assert x == [1, 2, 3]


def test_handle_restart_check_multi_objective_multi():
    from enn.turbo.config.morbo_tr_config import RescalePolicyConfig
    from enn.turbo.config import Rescalarize
    from enn.turbo.morbo_trust_region import MorboTrustRegion

    rng = np.random.default_rng(42)
    config = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        rescale_policy=RescalePolicyConfig(rescalarize=Rescalarize.ON_PROPOSE),
    )
    tr_state = MorboTrustRegion(config=config, num_dim=2, rng=rng)
    x = [1, 2, 3]
    y = [4, 5, 6]
    yvar = [0.1, 0.2, 0.3]
    cleared, idx = handle_restart_check_multi_objective(
        tr_state, x, y, yvar, init_idx=5
    )
    assert cleared is True
    assert idx == 0
    assert x == []


def test_estimate_y_passthrough_1d():
    y = np.array([1.0, 2.0, 3.0])
    result = estimate_y_passthrough(y)
    assert result.shape == (3, 1)


def test_estimate_y_passthrough_2d():
    y = np.array([[1.0, 2.0], [3.0, 4.0]])
    result = estimate_y_passthrough(y)
    assert result.shape == (2, 2)



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING
import numpy as np
from ..sampling import draw_lhd
from .optimization_strategy import OptimizationStrategy

if TYPE_CHECKING:
    from ..optimizer import Optimizer
    from ..types import TellInputs


@dataclass
class LHDOnlyStrategy(OptimizationStrategy):
    _bounds: np.ndarray
    _rng: object

    @classmethod
    def create(cls, *, bounds: np.ndarray, rng: object) -> LHDOnlyStrategy:
        bounds = np.asarray(bounds, dtype=float)
        if bounds.ndim != 2 or bounds.shape[1] != 2:
            raise ValueError(f"bounds must be (d, 2), got {bounds.shape}")
        return cls(_bounds=bounds, _rng=rng)

    def ask(self, opt: Optimizer, num_arms: int) -> np.ndarray:
        return draw_lhd(bounds=self._bounds, num_arms=num_arms, rng=opt._rng)

    def init_progress(self) -> tuple[int, int] | None:
        return None

    def tell(
        self, opt: Optimizer, inputs: TellInputs, *, x_unit: np.ndarray
    ) -> np.ndarray:
        del x_unit
        opt._y_tr_list = inputs.y.tolist()
        return np.asarray(inputs.y, dtype=float)



from __future__ import annotations
from typing import TYPE_CHECKING, Any
import numpy as np

if TYPE_CHECKING:
    from numpy.random import Generator
    from .protocols import Surrogate


class RandomAcqOptimizer:
    def select(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        surrogate: Surrogate,
        rng: Generator,
        *,
        tr_state: Any | None = None,
    ) -> np.ndarray:
        idx = rng.choice(len(x_cand), size=num_arms, replace=False)
        return x_cand[idx]



from __future__ import annotations
import numpy as np
from enn.turbo.proposal import mk_enn


def test_mk_enn_empty_returns_none():
    model, params = mk_enn([], [], k=3)
    assert model is None and params is None


def test_mk_enn_builds_model_and_params():
    rng = np.random.default_rng(0)
    x_obs = np.array([[0.0, 0.0], [1.0, 1.0], [0.25, 0.75]])
    y_obs = np.array([0.0, 1.0, 0.2])
    yvar_obs = np.array([0.1, 0.1, 0.1])
    model, params = mk_enn(
        x_obs,
        y_obs,
        k=3,
        yvar_obs=yvar_obs,
        rng=rng,
    )
    assert model is not None and params is not None
    assert params.k_num_neighbors == 3



from __future__ import annotations
import numpy as np
from enn.turbo.morbo_trust_region import MorboTrustRegion
from enn.turbo.config.morbo_tr_config import MorboTRConfig, MultiObjectiveConfig


def test_morbo_resample_weights():
    rng = np.random.default_rng(42)
    config = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2))
    tr = MorboTrustRegion(config, num_dim=3, rng=rng)
    weights_before = tr.weights.copy()
    tr.resample_weights(rng)
    weights_after = tr.weights.copy()
    assert not np.allclose(weights_before, weights_after)


def test_morbo_restart_resamples():
    from enn.turbo.config import Rescalarize
    from enn.turbo.config.morbo_tr_config import RescalePolicyConfig

    rng = np.random.default_rng(42)
    config = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        rescale_policy=RescalePolicyConfig(rescalarize=Rescalarize.ON_RESTART),
    )
    tr = MorboTrustRegion(config, num_dim=3, rng=rng)
    weights_before = tr.weights.copy()
    tr.restart(rng)
    weights_after = tr.weights.copy()
    assert not np.allclose(weights_before, weights_after)



from __future__ import annotations

import numpy as np
import pytest

from enn.turbo.turbo_utils import (
    ScalarIncumbentMixin,
    get_incumbent_index,
    get_scalar_incumbent_value,
    get_single_incumbent_index,
)


class MockSelector:
    def __init__(self, noise_aware=False):
        self.noise_aware = noise_aware

    def select(self, y, mu, rng):
        return np.argmax(mu if self.noise_aware else y)


def test_get_single_incumbent_index():
    selector = MockSelector()
    y = np.array([1.0, 3.0, 2.0])
    rng = np.random.default_rng(0)
    idx = get_single_incumbent_index(selector, y, rng)
    assert np.array_equal(idx, [1])
    assert get_single_incumbent_index(selector, np.array([]), rng).size == 0


def test_get_incumbent_index():
    selector = MockSelector()
    y = np.array([1.0, 3.0, 2.0])
    rng = np.random.default_rng(0)
    assert get_incumbent_index(selector, y, rng) == 1
    with pytest.raises(ValueError, match="empty"):
        get_incumbent_index(selector, np.array([]), rng)


def test_get_scalar_incumbent_value():
    selector = MockSelector(noise_aware=True)
    y = np.array([1.0, 3.0, 2.0])
    mu = np.array([1.5, 2.5, 2.1])
    rng = np.random.default_rng(0)
    val = get_scalar_incumbent_value(selector, y, rng, mu_obs=mu)
    assert val[0] == 2.5
    assert get_scalar_incumbent_value(selector, np.array([]), rng).size == 0


def test_scalar_incumbent_mixin():
    class TestMixin(ScalarIncumbentMixin):
        def __init__(self, selector):
            self.incumbent_selector = selector

    selector = MockSelector()
    mixin = TestMixin(selector)
    y = np.array([1.0, 3.0, 2.0])
    rng = np.random.default_rng(0)
    assert mixin.get_incumbent_index(y, rng) == 1
    assert mixin.get_incumbent_value(y, rng)[0] == 3.0



from __future__ import annotations

import pytest

from enn.turbo.config import Rescalarize
from enn.turbo.config.turbo_tr_config import TRLengthConfig
from enn.turbo.config import (
    AcqType,
    CandidateGenConfig,
    CandidateRV,
    DrawAcquisitionConfig,
    ENNFitConfig,
    ENNSurrogateConfig,
    GPSurrogateConfig,
    HnROptimizerConfig,
    HybridInit,
    InitConfig,
    LHDOnlyInit,
    MorboTRConfig,
    MultiObjectiveConfig,
    NDSOptimizerConfig,
    NoSurrogateConfig,
    NoTRConfig,
    OptimizerConfig,
    ParetoAcquisitionConfig,
    RAASPOptimizerConfig,
    RandomAcquisitionConfig,
    RescalePolicyConfig,
    TurboTRConfig,
    UCBAcquisitionConfig,
    lhd_only_config,
    turbo_enn_config,
    turbo_one_config,
    turbo_zero_config,
)


def test_turbo_tr_config_defaults():
    cfg = TurboTRConfig()
    assert cfg.length_init == 0.8
    assert cfg.length_min == 0.5**7
    assert cfg.length_max == 1.6


def test_turbo_tr_config_custom():
    cfg = TurboTRConfig(
        length=TRLengthConfig(length_init=0.5, length_min=0.01, length_max=2.0)
    )
    assert cfg.length_init == 0.5
    assert cfg.length_min == 0.01
    assert cfg.length_max == 2.0


def test_turbo_tr_config_invalid():
    with pytest.raises(ValueError, match="length_init must be > 0"):
        TRLengthConfig(length_init=0)
    with pytest.raises(ValueError, match="length_min must be < length_max"):
        TRLengthConfig(length_min=1.0, length_max=0.5)


def test_turbo_tr_config_invalid_length_init_exceeds_max():
    with pytest.raises(ValueError, match="length_init must be <= length_max"):
        TRLengthConfig(length_init=2.0, length_max=1.0)


def test_turbo_tr_config_invalid_length_min_exceeds_init():
    with pytest.raises(ValueError, match="length_min must be <= length_init"):
        TRLengthConfig(length_init=0.05, length_min=0.1)


def test_morbo_tr_config():
    cfg = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=3))
    assert cfg.num_metrics == 3
    assert cfg.alpha == 0.05


def test_morbo_tr_config_invalid():
    with pytest.raises(ValueError, match="num_metrics must be >= 2"):
        MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=1))


def test_morbo_tr_config_custom_alpha():
    cfg = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2, alpha=0.1))
    assert cfg.alpha == 0.1


def test_morbo_tr_config_invalid_alpha():
    with pytest.raises(ValueError, match="alpha must be > 0"):
        MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2, alpha=0))


def test_multi_objective_config_defaults():
    cfg = MultiObjectiveConfig(num_metrics=2)
    assert cfg.num_metrics == 2
    assert cfg.alpha == 0.05


def test_multi_objective_config_custom():
    cfg = MultiObjectiveConfig(num_metrics=3, alpha=0.1)
    assert cfg.num_metrics == 3
    assert cfg.alpha == 0.1


def test_morbo_tr_config_length_defaults():
    cfg = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2))
    assert cfg.length_init == 0.8
    assert cfg.length_min == 0.5**7
    assert cfg.length_max == 1.6


def test_morbo_tr_config_custom_length():
    cfg = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        length=TRLengthConfig(
            length_init=0.5,
            length_min=0.01,
            length_max=2.0,
        ),
    )
    assert cfg.length_init == 0.5
    assert cfg.length_min == 0.01
    assert cfg.length_max == 2.0


def test_morbo_tr_config_rescalarize_default():
    cfg = MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2))
    assert cfg.rescalarize == Rescalarize.ON_PROPOSE


def test_morbo_tr_config_rescalarize_custom():
    cfg = MorboTRConfig(
        multi_objective=MultiObjectiveConfig(num_metrics=2),
        rescale_policy=RescalePolicyConfig(rescalarize=Rescalarize.ON_RESTART),
    )
    assert cfg.rescalarize == Rescalarize.ON_RESTART


def test_rescale_policy_config_default():
    cfg = RescalePolicyConfig()
    assert cfg.rescalarize == Rescalarize.ON_PROPOSE


def test_rescale_policy_config_custom():
    cfg = RescalePolicyConfig(rescalarize=Rescalarize.ON_RESTART)
    assert cfg.rescalarize == Rescalarize.ON_RESTART


def test_no_tr_config():
    cfg = NoTRConfig()
    assert cfg is not None


def test_candidate_gen_config_defaults():
    cfg = CandidateGenConfig()
    assert cfg.candidate_rv == CandidateRV.SOBOL
    assert callable(cfg.num_candidates)
    assert cfg.num_candidates(num_dim=1, num_arms=1) == 100
    assert cfg.num_candidates(num_dim=100, num_arms=1) == 5000


def test_candidate_gen_config_uniform():
    cfg = CandidateGenConfig(
        candidate_rv=CandidateRV.UNIFORM,
        num_candidates=lambda *, num_dim, num_arms: 100,
    )
    assert cfg.candidate_rv == CandidateRV.UNIFORM
    assert callable(cfg.num_candidates)
    assert cfg.num_candidates(num_dim=3, num_arms=7) == 100


def test_candidate_gen_config_invalid_rv():
    with pytest.raises(ValueError, match="candidate_rv must be"):
        CandidateGenConfig(candidate_rv="invalid")


def test_candidate_gen_config_invalid_num_candidates():
    with pytest.raises(ValueError, match="num_candidates must be > 0"):
        CandidateGenConfig(num_candidates=lambda *, num_dim, num_arms: 0)


def test_candidate_gen_config_num_candidates_per_arms():
    cfg = CandidateGenConfig(num_candidates=lambda *, num_dim, num_arms: 100 * num_arms)
    assert cfg.num_candidates(num_dim=3, num_arms=7) == 700


def test_init_config_defaults():
    cfg = InitConfig()
    assert isinstance(cfg.init_strategy, HybridInit)
    assert cfg.num_init is None


def test_init_config_lhd_only():
    cfg = InitConfig(init_strategy=LHDOnlyInit(), num_init=20)
    assert isinstance(cfg.init_strategy, LHDOnlyInit)
    assert cfg.num_init == 20


def test_init_config_invalid_strategy():
    with pytest.raises(ValueError, match="init_strategy must be"):
        InitConfig(init_strategy="invalid")


def test_init_config_invalid_num_init():
    with pytest.raises(ValueError, match="num_init must be > 0"):
        InitConfig(num_init=0)


def test_no_surrogate_config():
    cfg = NoSurrogateConfig()
    assert cfg is not None


def test_gp_surrogate_config():
    cfg = GPSurrogateConfig()
    assert cfg is not None


def test_enn_surrogate_config_defaults():
    cfg = ENNSurrogateConfig()
    assert cfg.k is None
    assert cfg.num_fit_samples is None
    assert cfg.num_fit_candidates is None
    assert cfg.scale_x is False


def test_enn_surrogate_config_with_values():
    cfg = ENNSurrogateConfig(k=10, fit=ENNFitConfig(num_fit_samples=50), scale_x=True)
    assert cfg.k == 10
    assert cfg.num_fit_samples == 50
    assert cfg.scale_x is True


def test_enn_fit_config_defaults():
    cfg = ENNFitConfig()
    assert cfg.num_fit_samples is None
    assert cfg.num_fit_candidates is None


def test_enn_fit_config_custom():
    cfg = ENNFitConfig(num_fit_samples=50, num_fit_candidates=100)
    assert cfg.num_fit_samples == 50
    assert cfg.num_fit_candidates == 100


def test_enn_fit_config_invalid_num_fit_samples():
    with pytest.raises(ValueError, match="num_fit_samples must be > 0"):
        ENNFitConfig(num_fit_samples=0)


def test_enn_fit_config_invalid_num_fit_candidates():
    with pytest.raises(ValueError, match="num_fit_candidates must be > 0"):
        ENNFitConfig(num_fit_candidates=0)


def test_enn_surrogate_config_num_fit_candidates_custom():
    cfg = ENNSurrogateConfig(fit=ENNFitConfig(num_fit_candidates=100))
    assert cfg.num_fit_candidates == 100


def test_ucb_acquisition_config():
    cfg = UCBAcquisitionConfig()
    assert cfg is not None


def test_draw_acquisition_config():
    cfg = DrawAcquisitionConfig()
    assert cfg is not None


def test_pareto_acquisition_config():
    cfg = ParetoAcquisitionConfig()
    assert cfg is not None


def test_random_acquisition_config():
    cfg = RandomAcquisitionConfig()
    assert cfg is not None


def test_raasp_optimizer_config():
    cfg = RAASPOptimizerConfig()
    assert cfg is not None


def test_nds_optimizer_config():
    cfg = NDSOptimizerConfig()
    assert cfg is not None


def test_hnr_optimizer_config():
    cfg = HnROptimizerConfig()
    assert cfg is not None


def test_optimizer_config_defaults():
    cfg = OptimizerConfig()
    assert isinstance(cfg.trust_region, TurboTRConfig)
    assert cfg.candidate_rv == CandidateRV.SOBOL
    assert cfg.init.num_init is None
    assert isinstance(cfg.surrogate, NoSurrogateConfig)


def test_optimizer_config_lhd_only_requires_no_surrogate():
    with pytest.raises(
        ValueError, match="init_strategy='lhd_only' requires NoSurrogateConfig"
    ):
        OptimizerConfig(
            init=InitConfig(init_strategy=LHDOnlyInit()),
            surrogate=GPSurrogateConfig(),
        )
    config = OptimizerConfig(
        init=InitConfig(init_strategy=LHDOnlyInit()),
        trust_region=MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2)),
        surrogate=NoSurrogateConfig(),
    )
    assert config.num_metrics == 2


def test_optimizer_config_no_surrogate_rejects_draw():
    with pytest.raises(
        ValueError,
        match="NoSurrogateConfig is not compatible with DrawAcquisitionConfig",
    ):
        OptimizerConfig(
            surrogate=NoSurrogateConfig(),
            acquisition=DrawAcquisitionConfig(),
        )


def test_optimizer_config_no_surrogate_rejects_ucb():
    with pytest.raises(
        ValueError,
        match="NoSurrogateConfig is not compatible with UCBAcquisitionConfig",
    ):
        OptimizerConfig(
            surrogate=NoSurrogateConfig(),
            acquisition=UCBAcquisitionConfig(),
        )


def test_optimizer_config_pareto_requires_nds():
    with pytest.raises(
        ValueError, match="ParetoAcquisitionConfig requires NDSOptimizerConfig"
    ):
        OptimizerConfig(
            acquisition=ParetoAcquisitionConfig(),
            acq_optimizer=RAASPOptimizerConfig(),
        )


def test_optimizer_config_hnr_incompatible_with_pareto():
    with pytest.raises(
        ValueError, match="ParetoAcquisitionConfig requires NDSOptimizerConfig"
    ):
        OptimizerConfig(
            acquisition=ParetoAcquisitionConfig(),
            acq_optimizer=HnROptimizerConfig(),
        )


def test_optimizer_config_gp_draw_hnr_nyi():
    with pytest.raises(NotImplementedError, match="not yet implemented"):
        OptimizerConfig(
            surrogate=GPSurrogateConfig(),
            acquisition=DrawAcquisitionConfig(),
            acq_optimizer=HnROptimizerConfig(),
        )


def test_optimizer_config_enn_ucb_hnr_valid():
    cfg = OptimizerConfig(
        surrogate=ENNSurrogateConfig(fit=ENNFitConfig(num_fit_samples=50)),
        acquisition=UCBAcquisitionConfig(),
        acq_optimizer=HnROptimizerConfig(),
    )
    assert isinstance(cfg.acq_optimizer, HnROptimizerConfig)
    assert isinstance(cfg.acquisition, UCBAcquisitionConfig)


def test_optimizer_config_enn_draw_hnr_valid():
    cfg = OptimizerConfig(
        surrogate=ENNSurrogateConfig(fit=ENNFitConfig(num_fit_samples=50)),
        acquisition=DrawAcquisitionConfig(),
        acq_optimizer=HnROptimizerConfig(),
    )
    assert isinstance(cfg.acq_optimizer, HnROptimizerConfig)
    assert isinstance(cfg.acquisition, DrawAcquisitionConfig)


def test_optimizer_config_gp_ucb_hnr_valid():
    cfg = OptimizerConfig(
        surrogate=GPSurrogateConfig(),
        acquisition=UCBAcquisitionConfig(),
        acq_optimizer=HnROptimizerConfig(),
    )
    assert isinstance(cfg.acq_optimizer, HnROptimizerConfig)
    assert isinstance(cfg.acquisition, UCBAcquisitionConfig)


def test_turbo_one_config_factory():
    cfg = turbo_one_config()
    assert isinstance(cfg.trust_region, TurboTRConfig)
    assert isinstance(cfg.surrogate, GPSurrogateConfig)
    assert isinstance(cfg.acquisition, DrawAcquisitionConfig)


def test_turbo_zero_config_factory():
    cfg = turbo_zero_config()
    assert isinstance(cfg.trust_region, TurboTRConfig)
    assert isinstance(cfg.surrogate, NoSurrogateConfig)
    assert isinstance(cfg.acquisition, RandomAcquisitionConfig)


def test_turbo_enn_config_factory_pareto():
    cfg = turbo_enn_config(acq_type=AcqType.PARETO)
    assert isinstance(cfg.surrogate, ENNSurrogateConfig)
    assert isinstance(cfg.acquisition, ParetoAcquisitionConfig)
    assert isinstance(cfg.acq_optimizer, NDSOptimizerConfig)


def test_turbo_enn_config_factory_ucb():
    cfg = turbo_enn_config(
        acq_type=AcqType.UCB,
        enn=ENNSurrogateConfig(fit=ENNFitConfig(num_fit_samples=50)),
    )
    assert isinstance(cfg.surrogate, ENNSurrogateConfig)
    assert isinstance(cfg.acquisition, UCBAcquisitionConfig)
    assert isinstance(cfg.acq_optimizer, RAASPOptimizerConfig)


def test_turbo_enn_config_factory_thompson():
    cfg = turbo_enn_config(
        acq_type=AcqType.THOMPSON,
        enn=ENNSurrogateConfig(fit=ENNFitConfig(num_fit_samples=50)),
    )
    assert isinstance(cfg.surrogate, ENNSurrogateConfig)
    assert isinstance(cfg.acquisition, DrawAcquisitionConfig)


def test_turbo_enn_config_requires_num_fit_samples_for_non_pareto():
    with pytest.raises(ValueError, match="num_fit_samples required"):
        turbo_enn_config(acq_type=AcqType.UCB)


def test_lhd_only_config_factory():
    cfg = lhd_only_config()
    assert isinstance(cfg.trust_region, NoTRConfig)
    assert isinstance(cfg.init.init_strategy, LHDOnlyInit)
    assert isinstance(cfg.surrogate, NoSurrogateConfig)


def test_optimizer_config_properties():
    cfg = turbo_enn_config(
        enn=ENNSurrogateConfig(k=15),
        candidates=CandidateGenConfig(num_candidates=lambda *, num_dim, num_arms: 200),
        num_init=10,
    )
    assert cfg.surrogate.k == 15
    assert cfg.num_candidates(num_dim=3, num_arms=7) == 200
    assert cfg.init.num_init == 10


def test_optimizer_config_num_metrics():
    cfg_turbo = OptimizerConfig(trust_region=TurboTRConfig())
    assert cfg_turbo.num_metrics is None
    cfg_morbo = OptimizerConfig(
        trust_region=MorboTRConfig(multi_objective=MultiObjectiveConfig(num_metrics=2))
    )
    assert cfg_morbo.num_metrics == 2
    cfg_none = OptimizerConfig(trust_region=NoTRConfig())
    assert cfg_none.num_metrics is None



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np


def normal_hash_batch_multi_seed(
    function_seeds: np.ndarray, data_indices: np.ndarray, num_metrics: int
) -> np.ndarray:
    import numpy as np
    from scipy.special import ndtri

    num_seeds = len(function_seeds)
    unique_indices, inverse = np.unique(data_indices, return_inverse=True)
    num_unique = len(unique_indices)
    seed_grid, idx_grid, metric_grid = np.meshgrid(
        function_seeds.astype(np.uint64),
        unique_indices.astype(np.uint64),
        np.arange(num_metrics, dtype=np.uint64),
        indexing="ij",
    )
    seed_flat = seed_grid.ravel()
    idx_flat = idx_grid.ravel()
    metric_flat = metric_grid.ravel()
    combined_seeds = (seed_flat * np.uint64(1_000_003) + idx_flat) * np.uint64(
        1_000_003
    ) + metric_flat
    uniform_vals = np.empty(len(combined_seeds), dtype=float)
    for i, seed in enumerate(combined_seeds):
        rng = np.random.Generator(np.random.Philox(int(seed)))
        uniform_vals[i] = rng.random()
    uniform_vals = np.clip(uniform_vals, 1e-10, 1.0 - 1e-10)
    normal_vals = ndtri(uniform_vals).reshape(num_seeds, num_unique, num_metrics)
    return normal_vals[:, inverse.ravel(), :].reshape(
        num_seeds, *data_indices.shape, num_metrics
    )


def normal_hash_batch_multi_seed_fast(
    function_seeds: np.ndarray, data_indices: np.ndarray, num_metrics: int
) -> np.ndarray:
    import numpy as np

    function_seeds = np.asarray(function_seeds, dtype=np.int64)
    data_indices = np.asarray(data_indices)
    if num_metrics <= 0:
        raise ValueError(num_metrics)
    num_seeds = len(function_seeds)
    unique_indices, inverse = np.unique(data_indices, return_inverse=True)

    def _splitmix64(x: np.ndarray) -> np.ndarray:
        with np.errstate(over="ignore"):
            x = x + np.uint64(0x9E3779B97F4A7C15)
            z = x
            z = (z ^ (z >> np.uint64(30))) * np.uint64(0xBF58476D1CE4E5B9)
            z = (z ^ (z >> np.uint64(27))) * np.uint64(0x94D049BB133111EB)
            z = z ^ (z >> np.uint64(31))
            return z

    seeds_u64 = function_seeds.astype(np.uint64, copy=False)
    unique_u64 = unique_indices.astype(np.uint64, copy=False)
    metric_u64 = np.arange(num_metrics, dtype=np.uint64)
    normal_vals = np.empty((num_seeds, unique_indices.size, num_metrics), dtype=float)
    p = np.uint64(1_000_003)
    inv_2p53 = 1.0 / 9007199254740992.0
    for si, s in enumerate(seeds_u64):
        with np.errstate(over="ignore"):
            base = (s * p + unique_u64) * p
        combined = base[:, None] + metric_u64[None, :]
        r1 = _splitmix64(combined)
        r2 = _splitmix64(combined ^ np.uint64(0xD2B74407B1CE6E93))
        u1 = (r1 >> np.uint64(11)).astype(np.float64) * inv_2p53
        u2 = (r2 >> np.uint64(11)).astype(np.float64) * inv_2p53
        u1 = np.clip(u1, 1e-12, 1.0 - 1e-12)
        normal_vals[si, :, :] = np.sqrt(-2.0 * np.log(u1)) * np.cos(2.0 * np.pi * u2)
    return normal_vals[:, inverse.ravel(), :].reshape(
        num_seeds, *data_indices.shape, num_metrics
    )



from __future__ import annotations

from dataclasses import dataclass
from .rescalarize import Rescalarize


@dataclass(frozen=True)
class RescalePolicyConfig:
    rescalarize: Rescalarize = Rescalarize.ON_PROPOSE



from __future__ import annotations
from typing import TYPE_CHECKING, Any
import numpy as np

if TYPE_CHECKING:
    from numpy.random import Generator
    from .protocols import Surrogate


class ThompsonAcqOptimizer:
    def select(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        surrogate: Surrogate,
        rng: Generator,
        *,
        tr_state: Any | None = None,
    ) -> np.ndarray:
        from ..turbo_utils import argmax_random_tie

        num_candidates = len(x_cand)
        samples = surrogate.sample(x_cand, num_arms, rng)
        assert samples.ndim == 3, f"samples.ndim={samples.ndim}, expected 3"
        assert samples.shape[0] == num_arms, (
            f"samples.shape[0]={samples.shape[0]}, expected num_arms={num_arms}"
        )
        assert samples.shape[1] == num_candidates, (
            f"samples.shape[1]={samples.shape[1]}, expected num_candidates={num_candidates}"
        )
        num_metrics = samples.shape[2]
        if tr_state is not None and hasattr(tr_state, "scalarize"):
            indices = []
            # Vectorize scalarization across all arms and candidates
            # samples is (num_arms, num_candidates, num_metrics)
            # Reshape to (num_arms * num_candidates, num_metrics) for scalarize
            flat_samples = samples.reshape(-1, num_metrics)
            flat_scores = tr_state.scalarize(flat_samples, clip=False)
            all_scores = flat_scores.reshape(num_arms, num_candidates)

            for i in range(num_arms):
                scores = all_scores[i].copy()
                for prev_idx in indices:
                    scores[prev_idx] = -np.inf
                idx = argmax_random_tie(scores, rng=rng)
                indices.append(idx)
            return x_cand[indices]
        else:
            arm_indices = [
                argmax_random_tie(samples[i, :, 0], rng=rng) for i in range(num_arms)
            ]
            return x_cand[arm_indices]



import numpy as np
import pytest
from scipy.stats import qmc
from enn.turbo.turbo_utils import generate_tr_candidates, generate_tr_candidates_fast
from enn.turbo.config import CandidateRV, RAASPDriver


def test_candidate_generation_statistical_properties():
    """
    Compares the statistical properties of generate_tr_candidates and generate_tr_candidates_fast.
    We check:
    1. Mean and variance of perturbed dimensions.
    2. Number of perturbed dimensions per candidate.
    3. Range of generated values.
    """
    num_dim = 100
    num_candidates = 1000
    num_pert = 20
    rng = np.random.default_rng(42)
    x_center = np.full(num_dim, 0.5)

    def compute_bounds(center, lengthscales=None):
        return np.zeros_like(center), np.ones_like(center)

    # --- 1. Test Uniform Perturbations ---
    orig_uniform = generate_tr_candidates(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.UNIFORM,
        sobol_engine=None,
        raasp_driver=RAASPDriver.ORIG,
        num_pert=num_pert,
    )

    fast_uniform = generate_tr_candidates(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.UNIFORM,
        sobol_engine=None,
        raasp_driver=RAASPDriver.FAST,
        num_pert=num_pert,
    )

    for name, cand in [
        ("Original Uniform", orig_uniform),
        ("Fast Uniform", fast_uniform),
    ]:
        # Check shape
        assert cand.shape == (num_candidates, num_dim)

        # Check that center is preserved in non-perturbed dimensions
        # Count how many elements are NOT equal to the center value
        perturbed_mask = ~np.isclose(cand, 0.5)
        per_row_pert = perturbed_mask.sum(axis=1)

        # Both should now follow the same distribution (Binomial)
        # Mean should be approx num_pert
        assert np.mean(per_row_pert) == pytest.approx(num_pert, abs=2.0)
        # Variance should be approx num_pert * (1 - num_pert/num_dim)
        expected_var = num_pert * (1 - num_pert / num_dim)
        assert np.var(per_row_pert) == pytest.approx(expected_var, abs=5.0)
        # Ensure at least one dimension is perturbed in every row
        assert np.all(per_row_pert >= 1)

        # Check that values are within [0, 1]
        assert np.all(cand >= 0.0)
        assert np.all(cand <= 1.0)

        # Check mean of perturbed values (should be ~0.5)
        perturbed_values = cand[perturbed_mask]
        assert np.mean(perturbed_values) == pytest.approx(0.5, abs=0.05)
        # Check variance of perturbed values (uniform [0,1] variance is 1/12 approx 0.0833)
        assert np.var(perturbed_values) == pytest.approx(0.0833, abs=0.01)

    # --- 2. Test Sobol Perturbations ---
    sobol_engine = qmc.Sobol(d=num_dim, scramble=True, seed=42)
    orig_sobol = generate_tr_candidates(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.SOBOL,
        sobol_engine=sobol_engine,
        raasp_driver=RAASPDriver.ORIG,
        num_pert=num_pert,
    )

    fast_sobol = generate_tr_candidates(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.SOBOL,
        sobol_engine=None,
        raasp_driver=RAASPDriver.FAST,
        num_pert=num_pert,
    )

    for name, cand in [("Original Sobol", orig_sobol), ("Fast Sobol", fast_sobol)]:
        perturbed_mask = ~np.isclose(cand, 0.5)
        perturbed_values = cand[perturbed_mask]

        # Sobol should also be roughly uniform in [0, 1]
        assert np.mean(perturbed_values) == pytest.approx(0.5, abs=0.05)
        assert np.var(perturbed_values) == pytest.approx(0.0833, abs=0.01)


def test_fast_candidates_various_inputs():
    """Test generate_tr_candidates_fast with different num_pert and dimensions."""
    num_dim = 50
    num_candidates = 100
    rng = np.random.default_rng(0)
    x_center = np.zeros(num_dim)

    def compute_bounds(center, lengthscales=None):
        return np.full_like(center, -1.0), np.full_like(center, 1.0)

    # num_pert = 1
    cand = generate_tr_candidates_fast(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.UNIFORM,
        num_pert=1,
    )
    perturbed_mask = ~np.isclose(cand, 0.0)
    # Mean should be 1, but individual rows vary (Binomial)
    assert np.mean(perturbed_mask.sum(axis=1)) == pytest.approx(1.0, abs=0.5)
    assert np.all(perturbed_mask.sum(axis=1) >= 1)

    # Test num_pert = num_dim
    cand = generate_tr_candidates_fast(
        compute_bounds,
        x_center,
        None,
        num_candidates,
        rng=rng,
        candidate_rv=CandidateRV.UNIFORM,
        num_pert=num_dim,
    )
    perturbed_mask = ~np.isclose(cand, 0.0)
    assert np.all(perturbed_mask.sum(axis=1) == num_dim)



from __future__ import annotations
from dataclasses import dataclass


@dataclass(frozen=True)
class ENNParams:
    k_num_neighbors: int
    epistemic_variance_scale: float
    aleatoric_variance_scale: float

    def __post_init__(self) -> None:
        import numpy as np

        k = int(self.k_num_neighbors)
        if k <= 0:
            raise ValueError(f"k_num_neighbors must be > 0, got {k}")
        epi_var_scale = float(self.epistemic_variance_scale)
        if not np.isfinite(epi_var_scale) or epi_var_scale < 0.0:
            raise ValueError(
                f"epistemic_variance_scale must be >= 0, got {epi_var_scale}"
            )
        ale_scale = float(self.aleatoric_variance_scale)
        if not np.isfinite(ale_scale) or ale_scale < 0.0:
            raise ValueError(f"aleatoric_variance_scale must be >= 0, got {ale_scale}")



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Protocol

from .morbo_tr_config import MorboTRConfig, MultiObjectiveConfig, RescalePolicyConfig
from .no_tr_config import NoTRConfig
from .turbo_tr_config import TRLengthConfig, TurboTRConfig

if TYPE_CHECKING:
    import numpy as np
    from numpy.random import Generator
    from ..components.protocols import TrustRegion
    from .candidate_rv import CandidateRV
    from ..strategies import OptimizationStrategy


class TrustRegionConfig(Protocol):
    def build(
        self,
        *,
        num_dim: int,
        rng: Generator,
        candidate_rv: CandidateRV,
    ) -> TrustRegion: ...


class InitStrategy(ABC):
    @abstractmethod
    def create_runtime_strategy(
        self,
        *,
        bounds: np.ndarray,
        rng: Generator,
        num_init: int | None,
    ) -> OptimizationStrategy: ...


__all__ = [
    "InitStrategy",
    "MorboTRConfig",
    "MultiObjectiveConfig",
    "NoTRConfig",
    "RescalePolicyConfig",
    "TRLengthConfig",
    "TrustRegionConfig",
    "TurboTRConfig",
]



from __future__ import annotations
import warnings
import numpy as np
import pytest
import torch
from gpytorch.constraints import Interval
from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import FixedNoiseGaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood
from enn.turbo.components.gp_surrogate import GPSurrogate
from enn.turbo.turbo_gp import TurboGP
from enn.turbo.turbo_gp_base import TurboGPBase
from enn.turbo.turbo_gp_fit import fit_gp
from enn.turbo.turbo_gp_noisy import TurboGPNoisy


def _make_turbo_gp_noisy(
    *,
    train_x,
    train_y,
    train_y_var,
    ard_dims: int,
    learn_additional_noise: bool = False,
):
    lengthscale_constraint = Interval(0.005, 2.0)
    outputscale_constraint = Interval(0.05, 20.0)
    return TurboGPNoisy(
        train_x=train_x,
        train_y=train_y,
        train_y_var=train_y_var,
        lengthscale_constraint=lengthscale_constraint,
        outputscale_constraint=outputscale_constraint,
        ard_dims=ard_dims,
        learn_additional_noise=learn_additional_noise,
    )


def test_fit_gp_returns_model_with_valid_data():
    num_obs, num_dim = 20, 3
    x_obs = np.random.default_rng(0).random((num_obs, num_dim))
    y_obs = (
        x_obs.sum(axis=1) + 0.1 * np.random.default_rng(1).standard_normal(num_obs)
    ).tolist()
    result = fit_gp(x_obs.tolist(), y_obs, num_dim, num_steps=10)
    assert result.model is not None and result.likelihood is not None
    assert (
        isinstance(result.y_mean, float)
        and isinstance(result.y_std, float)
        and result.y_std > 0.0
    )


def test_fit_gp_returns_none_with_empty_data_and_returns_model_with_single_obs():
    num_dim = 2
    result_empty = fit_gp([], [], num_dim, num_steps=10)
    assert result_empty.model is None and result_empty.likelihood is None
    assert result_empty.y_mean == 0.0 and result_empty.y_std == 1.0
    x_single = np.random.default_rng(0).random((1, num_dim))
    result_single = fit_gp(x_single.tolist(), [1.0], num_dim, num_steps=0)
    assert result_single.model is not None and result_single.likelihood is not None
    assert result_single.y_mean == 1.0 and result_single.y_std == 1.0


def test_fit_gp_with_y_var_list_uses_noisy_model():
    num_obs, num_dim = 20, 3
    rng = np.random.default_rng(42)
    x_obs = rng.random((num_obs, num_dim))
    y_obs = (x_obs.sum(axis=1) + 0.1 * rng.standard_normal(num_obs)).tolist()
    y_var = rng.uniform(0.01, 0.1, size=num_obs).tolist()
    result = fit_gp(x_obs.tolist(), y_obs, num_dim, yvar_obs_list=y_var, num_steps=10)
    assert result.model is not None and isinstance(result.model, TurboGPNoisy)
    assert result.likelihood is not None and result.y_std > 0.0


def test_fit_gp_with_y_var_list_asserts_length():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(0)
    x_obs = rng.random((num_obs, num_dim)).tolist()
    y_obs = rng.random(num_obs).tolist()
    y_var_wrong = rng.uniform(0.01, 0.1, size=num_obs - 2).tolist()
    with pytest.raises(ValueError):
        fit_gp(x_obs, y_obs, num_dim, yvar_obs_list=y_var_wrong, num_steps=5)


def test_turbo_gp_noisy_accepts_train_y_var():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(42)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.1 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.as_tensor(
        rng.uniform(0.01, 0.1, size=num_obs), dtype=torch.float64
    )
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    assert model is not None and isinstance(
        model.likelihood, FixedNoiseGaussianLikelihood
    )


def test_turbo_gp_noisy_forward_and_posterior():
    num_obs, num_dim = 15, 3
    rng = np.random.default_rng(123)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.05 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.full((num_obs,), 0.01, dtype=torch.float64)
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    model.eval()
    model.likelihood.eval()
    test_x = torch.as_tensor(rng.random((5, num_dim)), dtype=torch.float64)
    with torch.no_grad():
        forward_output = model.forward(test_x)
        posterior_output = model.posterior(test_x)
    assert isinstance(forward_output, MultivariateNormal) and isinstance(
        posterior_output, MultivariateNormal
    )
    assert forward_output.mean.shape == (5,) and posterior_output.mean.shape == (5,)


def test_turbo_gp_noisy_trains_successfully():
    num_obs, num_dim = 20, 2
    rng = np.random.default_rng(999)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.1 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    train_y_var = torch.as_tensor(
        rng.uniform(0.005, 0.05, size=num_obs), dtype=torch.float64
    )
    model = _make_turbo_gp_noisy(
        train_x=train_x, train_y=train_y, train_y_var=train_y_var, ard_dims=num_dim
    )
    model.train()
    model.likelihood.train()
    mll = ExactMarginalLogLikelihood(model.likelihood, model)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
    initial_loss = None
    for i in range(20):
        optimizer.zero_grad()
        output = model(train_x)
        loss = -mll(output, train_y)
        if i == 0:
            initial_loss = loss.item()
        loss.backward()
        optimizer.step()
    assert loss.item() <= initial_loss


def test_turbo_gp_noisy_with_zero_variance():
    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(42)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(train_x.sum(dim=1).numpy(), dtype=torch.float64)
    train_y_var = torch.zeros(num_obs, dtype=torch.float64)
    from gpytorch.utils.warnings import NumericalWarning

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"Very small noise values detected\..*",
            category=NumericalWarning,
        )
        model = _make_turbo_gp_noisy(
            train_x=train_x,
            train_y=train_y,
            train_y_var=train_y_var,
            ard_dims=num_dim,
            learn_additional_noise=True,
        )
    model.eval()
    model.likelihood.eval()
    test_x = torch.as_tensor(rng.random((3, num_dim)), dtype=torch.float64)
    with torch.no_grad():
        posterior = model.posterior(test_x)
    assert posterior.mean.shape == (3,)


def test_fit_gp_multi_output_can_trigger_non_scalar_backward_error():
    rng = np.random.default_rng(0)
    num_dim, num_metrics, n = 3, 2, 8
    x = rng.uniform(0.0, 1.0, size=(n, num_dim))
    y = rng.normal(size=(n, num_metrics))
    result = fit_gp(x.tolist(), y.tolist(), num_dim, num_steps=0)
    assert result.model is not None and result.likelihood is not None
    result.model.train()
    result.likelihood.train()
    train_x = result.model.train_inputs[0]
    train_y = result.model.train_targets
    output = result.model(train_x)
    mll = ExactMarginalLogLikelihood(result.likelihood, result.model)
    loss = -mll(output, train_y)
    assert tuple(loss.shape) == (num_metrics,)
    with pytest.raises(
        RuntimeError, match="grad can be implicitly created only for scalar outputs"
    ):
        loss.backward()


def test_fit_gp_multi_output_trains_without_scalar_backward_error():
    rng = np.random.default_rng(0)
    num_dim, num_metrics, n = 3, 2, 8
    x = rng.uniform(0.0, 1.0, size=(n, num_dim))
    y = rng.normal(size=(n, num_metrics))
    result = fit_gp(x.tolist(), y.tolist(), num_dim, num_steps=2)
    assert result.model is not None and result.likelihood is not None
    assert result.y_mean.shape == (num_metrics,) and result.y_std.shape == (
        num_metrics,
    )


def test_gp_surrogate_predict_shapes_multi_output():
    rng = np.random.default_rng(0)
    num_dim, num_metrics, n = 3, 2, 6
    x = rng.uniform(0.0, 1.0, size=(n, num_dim))
    y = x @ rng.normal(size=(num_dim, num_metrics))
    surrogate = GPSurrogate()
    surrogate.fit(x, y, num_steps=2)
    posterior = surrogate.predict(x)
    mu = np.asarray(posterior.mu, dtype=float)
    sigma = np.asarray(posterior.sigma, dtype=float)
    assert mu.shape == (n, num_metrics)
    assert sigma.shape == (n, num_metrics)


def test_turbo_gp_init_and_forward():
    from gpytorch.likelihoods import GaussianLikelihood

    num_obs, num_dim = 10, 2
    rng = np.random.default_rng(42)
    train_x = torch.as_tensor(rng.random((num_obs, num_dim)), dtype=torch.float64)
    train_y = torch.as_tensor(
        train_x.sum(dim=1).numpy() + 0.1 * rng.standard_normal(num_obs),
        dtype=torch.float64,
    )
    lengthscale_constraint = Interval(0.005, 2.0)
    outputscale_constraint = Interval(0.05, 20.0)
    likelihood = GaussianLikelihood()
    model = TurboGP(
        train_x=train_x,
        train_y=train_y,
        likelihood=likelihood,
        lengthscale_constraint=lengthscale_constraint,
        outputscale_constraint=outputscale_constraint,
        ard_dims=num_dim,
    )
    assert model is not None
    assert isinstance(model, TurboGPBase)
    model.eval()
    likelihood.eval()
    test_x = torch.as_tensor(rng.random((5, num_dim)), dtype=torch.float64)
    with torch.no_grad():
        output = model.forward(test_x)
    assert isinstance(output, MultivariateNormal)
    assert output.mean.shape == (5,)


def test_turbo_gp_base_is_subclassed():
    assert TurboGPBase is not None
    assert issubclass(TurboGP, TurboGPBase)



from .draw_internals import DrawInternals
from .neighbor_data import NeighborData
from .weighted_stats import WeightedStats

_DrawInternals = DrawInternals
_NeighborData = NeighborData
_WeightedStats = WeightedStats

__all__ = [
    "DrawInternals",
    "NeighborData",
    "WeightedStats",
    "_DrawInternals",
    "_NeighborData",
    "_WeightedStats",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import numpy as np


def _use_faiss() -> bool:
    import os
    import sys

    if sys.platform == "darwin":
        return os.environ.get("ENN_USE_FAISS", "0") == "1"
    return True


class ENNIndex:
    def __init__(
        self,
        train_x_scaled: np.ndarray,
        num_dim: int,
        x_scale: np.ndarray,
        scale_x: bool,
        driver: Any = None,
    ) -> None:
        from enn.turbo.config.enn_index_driver import ENNIndexDriver

        if driver is None:
            driver = ENNIndexDriver.FLAT
        self._train_x_scaled = train_x_scaled
        self._num_dim = num_dim
        self._x_scale = x_scale
        self._scale_x = scale_x
        self._driver = driver
        self._index: Any | None = None
        self._build_index()

    def _build_index(self) -> None:
        import numpy as np

        from enn.turbo.config.enn_index_driver import ENNIndexDriver

        if len(self._train_x_scaled) == 0:
            return
        if _use_faiss():
            import faiss

            x_f32 = self._train_x_scaled.astype(np.float32, copy=False)
            if self._driver == ENNIndexDriver.FLAT:
                index = faiss.IndexFlatL2(self._num_dim)
            elif self._driver == ENNIndexDriver.HNSW:
                index = faiss.IndexHNSWFlat(self._num_dim, 32)
            else:
                raise ValueError(f"Unknown driver: {self._driver}")
            index.add(x_f32)
            self._index = index
        else:
            self._index = None

    def add(self, x: np.ndarray) -> None:
        import numpy as np

        x = np.asarray(x, dtype=float)
        if x.ndim != 2 or x.shape[1] != self._num_dim:
            raise ValueError(x.shape)
        x_scaled = x / self._x_scale if self._scale_x else x
        x_f32 = x_scaled.astype(np.float32, copy=False)
        self._train_x_scaled = np.concatenate([self._train_x_scaled, x_f32], axis=0)
        if self._index is not None:
            self._index.add(x_f32)

    def search(
        self,
        x: np.ndarray,
        *,
        search_k: int,
        exclude_nearest: bool,
    ) -> tuple[np.ndarray, np.ndarray]:
        import numpy as np

        search_k = int(search_k)
        if search_k <= 0:
            raise ValueError(search_k)
        x = np.asarray(x, dtype=float)
        if x.ndim != 2 or x.shape[1] != self._num_dim:
            raise ValueError(x.shape)
        x_scaled = x / self._x_scale if self._scale_x else x
        x_f32 = x_scaled.astype(np.float32, copy=False)
        if self._index is not None:
            dist2s_full, idx_full = self._index.search(x_f32, search_k)
            dist2s_full = dist2s_full.astype(float)
            idx_full = idx_full.astype(int)
        else:
            n_query = x_f32.shape[0]
            n_train = len(self._train_x_scaled)
            k = min(search_k, n_train)
            dist2s_full = np.full((n_query, search_k), np.inf, dtype=float)
            idx_full = np.full((n_query, search_k), -1, dtype=int)
            if k:
                x2 = np.sum(x_f32**2, axis=1, keepdims=True)
                y2 = np.sum(self._train_x_scaled**2, axis=1, keepdims=True).T
                d2 = x2 + y2 - 2.0 * (x_f32 @ self._train_x_scaled.T)
                part = np.argpartition(d2, kth=k - 1, axis=1)[:, :k]
                rows = np.arange(n_query)[:, None]
                d2_part = d2[rows, part]
                order = np.argsort(d2_part, axis=1)
                part_sorted = part[rows, order]
                d2_sorted = d2_part[rows, order]
                idx_full[:, :k] = part_sorted.astype(int)
                dist2s_full[:, :k] = d2_sorted.astype(float)
        if exclude_nearest:
            dist2s_full = dist2s_full[:, 1:]
            idx_full = idx_full[:, 1:]
        return dist2s_full, idx_full



from __future__ import annotations

import importlib
from typing import Any


def lazy_getattr(
    *,
    name: str,
    module_name: str,
    package: str,
    mapping: dict[str, tuple[str, str]],
    extra: str,
) -> Any:
    spec = mapping.get(name)
    if spec is None:
        raise AttributeError(f"module {module_name!r} has no attribute {name!r}")
    rel_module, attr_name = spec
    try:
        module = importlib.import_module(rel_module, package)
        return getattr(module, attr_name)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(f"{e}. Install extras via {extra}.") from e



import numpy as np
from enn.benchmarks import Ackley, DoubleAckley


def test_ackley_1d_input():
    rng = np.random.default_rng(42)
    ackley = Ackley(noise=0.0, rng=rng)
    x = np.zeros(5)
    result = ackley(x)
    assert result.shape == (1,)


def test_ackley_2d_input():
    rng = np.random.default_rng(42)
    ackley = Ackley(noise=0.0, rng=rng)
    x = np.zeros((10, 5))
    result = ackley(x)
    assert result.shape == (10,)


def test_ackley_optimum_near_one():
    rng = np.random.default_rng(42)
    ackley = Ackley(noise=0.0, rng=rng)
    x_opt = np.ones((1, 5))
    x_off = np.zeros((1, 5))
    y_opt = ackley(x_opt)[0]
    y_off = ackley(x_off)[0]
    assert y_opt > y_off


def test_ackley_deterministic_with_same_rng():
    rng1 = np.random.default_rng(42)
    rng2 = np.random.default_rng(42)
    ackley1 = Ackley(noise=0.1, rng=rng1)
    ackley2 = Ackley(noise=0.1, rng=rng2)
    x = np.random.default_rng(0).random((5, 3))
    np.testing.assert_array_equal(ackley1(x), ackley2(x))


def test_ackley_bounds():
    rng = np.random.default_rng(42)
    ackley = Ackley(noise=0.0, rng=rng)
    assert ackley.bounds == [-32.768, 32.768]


def test_double_ackley_shape():
    rng = np.random.default_rng(42)
    obj = DoubleAckley(noise=0.0, rng=rng)
    x = np.zeros((10, 6))
    result = obj(x)
    assert result.shape == (10, 2)


def test_double_ackley_requires_even_dims():
    rng = np.random.default_rng(42)
    obj = DoubleAckley(noise=0.0, rng=rng)
    x = np.zeros((5, 5))
    try:
        obj(x)
        assert False, "Should have raised ValueError"
    except ValueError:
        pass


def test_double_ackley_deterministic():
    rng1 = np.random.default_rng(42)
    rng2 = np.random.default_rng(42)
    obj1 = DoubleAckley(noise=0.1, rng=rng1)
    obj2 = DoubleAckley(noise=0.1, rng=rng2)
    x = np.random.default_rng(0).random((5, 6))
    np.testing.assert_array_equal(obj1(x), obj2(x))



from __future__ import annotations
from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from .candidate_gen_config import CandidateGenConfig
from .init_config import InitConfig
from .surrogate import NoSurrogateConfig, SurrogateConfig
from .trust_region import TrustRegionConfig, TurboTRConfig
from .observation_history_config import ObservationHistoryConfig

if TYPE_CHECKING:
    from .acquisition import AcqOptimizerConfig, AcquisitionConfig
    from .candidate_rv import CandidateRV
    from .raasp_driver import RAASPDriver


def _default_acquisition():
    from .acquisition import RandomAcquisitionConfig

    return RandomAcquisitionConfig()


def _default_acq_optimizer():
    from .raasp_optimizer_config import RAASPOptimizerConfig

    return RAASPOptimizerConfig()


@dataclass(frozen=True)
class OptimizerConfig:
    trust_region: TrustRegionConfig = TurboTRConfig()
    candidates: CandidateGenConfig = CandidateGenConfig()
    init: InitConfig = InitConfig()
    surrogate: SurrogateConfig = NoSurrogateConfig()
    acquisition: AcquisitionConfig = field(default_factory=_default_acquisition)
    acq_optimizer: AcqOptimizerConfig = field(default_factory=_default_acq_optimizer)
    observation_history: ObservationHistoryConfig = ObservationHistoryConfig()

    def __post_init__(self) -> None:
        from .validation import validate_optimizer_config

        validate_optimizer_config(self)

    @property
    def num_metrics(self) -> int | None:
        from .morbo_tr_config import MorboTRConfig

        if isinstance(self.trust_region, MorboTRConfig):
            return self.trust_region.num_metrics
        return None

    @property
    def candidate_rv(self) -> CandidateRV:
        return self.candidates.candidate_rv

    @property
    def raasp_driver(self) -> RAASPDriver:
        return self.candidates.raasp_driver

    @property
    def num_candidates(self):
        return self.candidates.num_candidates

    @property
    def trailing_obs(self) -> int | None:
        return self.observation_history.trailing_obs



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np


@dataclass(frozen=True)
class Neighbors:
    dist2: np.ndarray
    ids: np.ndarray
    y: np.ndarray
    yvar: np.ndarray | None



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


@dataclass(frozen=True)
class ENNFitConfig:
    num_fit_samples: int | None = None
    num_fit_candidates: int | None = None

    def __post_init__(self) -> None:
        if self.num_fit_samples is not None and self.num_fit_samples <= 0:
            raise ValueError(f"num_fit_samples must be > 0, got {self.num_fit_samples}")
        if self.num_fit_candidates is not None and self.num_fit_candidates <= 0:
            raise ValueError(
                f"num_fit_candidates must be > 0, got {self.num_fit_candidates}"
            )



from __future__ import annotations
from typing import TYPE_CHECKING, Any
import numpy as np

if TYPE_CHECKING:
    from numpy.random import Generator
    from .protocols import Surrogate


class UCBAcqOptimizer:
    def __init__(self, beta: float = 1.0) -> None:
        self._beta = beta

    def select(
        self,
        x_cand: np.ndarray,
        num_arms: int,
        surrogate: Surrogate,
        rng: Generator,
        *,
        tr_state: Any | None = None,
    ) -> np.ndarray:
        num_candidates = len(x_cand)
        posterior = surrogate.predict(x_cand)
        mu = posterior.mu
        sigma = posterior.sigma if posterior.sigma is not None else np.zeros_like(mu)
        assert mu.ndim == 2, f"mu.ndim={mu.ndim}, expected 2"
        assert mu.shape[0] == num_candidates, (
            f"mu.shape[0]={mu.shape[0]}, expected {num_candidates}"
        )
        num_metrics = mu.shape[1]
        if tr_state is not None and hasattr(tr_state, "scalarize"):
            ucb = mu + self._beta * sigma
            assert ucb.shape == (
                num_candidates,
                num_metrics,
            ), f"ucb.shape={ucb.shape}, expected ({num_candidates}, {num_metrics})"
            scores = tr_state.scalarize(ucb, clip=False)
            assert scores.shape == (num_candidates,), (
                f"scores.shape={scores.shape}, expected ({num_candidates},)"
            )
        else:
            scores = mu[:, 0] + self._beta * sigma[:, 0]
            assert scores.shape == (num_candidates,)
        shuffled_indices = rng.permutation(len(scores))
        shuffled_scores = scores[shuffled_indices]
        top_k_in_shuffled = np.argpartition(-shuffled_scores, num_arms - 1)[:num_arms]
        idx = shuffled_indices[top_k_in_shuffled]
        return x_cand[idx]



from __future__ import annotations
from typing import Any
from enn.enn.enn_util import standardize_y
from .types import GPDataPrep, GPFitResult


def _prepare_gp_data(
    x_obs_list: list, y_obs_list: list, yvar_obs_list: list | None
) -> GPDataPrep:
    import numpy as np
    import torch

    x = np.asarray(x_obs_list, dtype=float)
    y = np.asarray(y_obs_list, dtype=float)
    if y.ndim not in (1, 2):
        raise ValueError(y.shape)
    is_multi = y.ndim == 2 and y.shape[1] > 1
    if yvar_obs_list is not None:
        if len(yvar_obs_list) != len(y_obs_list):
            raise ValueError(
                f"yvar_obs_list length {len(yvar_obs_list)} != y_obs_list length {len(y_obs_list)}"
            )
        if is_multi:
            raise ValueError("yvar_obs_list not supported for multi-output GP")
    if is_multi:
        y_mean, y_std = y.mean(axis=0), y.std(axis=0)
        y_std = np.where(y_std < 1e-6, 1.0, y_std)
        z = (y - y_mean) / y_std
        train_y = torch.as_tensor(z.T, dtype=torch.float64)
    else:
        y_mean, y_std = standardize_y(y)
        z = (y - y_mean) / y_std
        train_y = torch.as_tensor(z, dtype=torch.float64)
    return GPDataPrep(
        train_x=torch.as_tensor(x, dtype=torch.float64),
        train_y=train_y,
        is_multi=is_multi,
        y_mean=y_mean,
        y_std=y_std,
        y_raw=y,
    )


def _build_gp_model(
    train_x: Any,
    train_y: Any,
    is_multi: bool,
    num_dim: int,
    *,
    yvar_obs_list: list | None,
    gp_y_std: Any,
    y: Any,
) -> tuple[Any, Any]:
    import numpy as np
    import torch
    from gpytorch.constraints import Interval
    from gpytorch.likelihoods import GaussianLikelihood
    from .turbo_gp import TurboGP
    from .turbo_gp_noisy import TurboGPNoisy

    ls_constr, os_constr = Interval(0.005, 2.0), Interval(0.05, 20.0)
    if yvar_obs_list is not None:
        y_var = np.asarray(yvar_obs_list, dtype=float)
        train_y_var = torch.as_tensor(y_var / (gp_y_std**2), dtype=torch.float64)
        model = TurboGPNoisy(
            train_x=train_x,
            train_y=train_y,
            train_y_var=train_y_var,
            lengthscale_constraint=ls_constr,
            outputscale_constraint=os_constr,
            ard_dims=num_dim,
        ).to(dtype=train_x.dtype)
        return model, model.likelihood
    noise_constr = Interval(5e-4, 0.2)
    num_out = int(y.shape[1]) if is_multi else None
    if is_multi:
        likelihood = GaussianLikelihood(
            noise_constraint=noise_constr, batch_shape=torch.Size([num_out])
        ).to(dtype=train_y.dtype)
    else:
        likelihood = GaussianLikelihood(noise_constraint=noise_constr).to(
            dtype=train_y.dtype
        )
    model = TurboGP(
        train_x=train_x,
        train_y=train_y,
        likelihood=likelihood,
        lengthscale_constraint=ls_constr,
        outputscale_constraint=os_constr,
        ard_dims=num_dim,
    ).to(dtype=train_x.dtype)
    likelihood.noise = (
        torch.full((num_out,), 0.005, dtype=train_y.dtype)
        if is_multi
        else torch.tensor(0.005, dtype=train_y.dtype)
    )
    return model, likelihood


def _init_gp_hyperparams(
    model: Any, is_multi: bool, num_dim: int, num_out: int | None, dtype: Any
) -> None:
    import torch

    if is_multi:
        model.covar_module.outputscale = torch.ones(num_out, dtype=dtype)
        model.covar_module.base_kernel.lengthscale = torch.full(
            (num_out, 1, num_dim), 0.5, dtype=dtype
        )
    else:
        model.covar_module.outputscale = torch.tensor(1.0, dtype=dtype)
        model.covar_module.base_kernel.lengthscale = torch.full(
            (num_dim,), 0.5, dtype=dtype
        )


def _train_gp(
    model: Any, likelihood: Any, train_x: Any, train_y: Any, num_steps: int
) -> None:
    import torch
    from gpytorch.mlls import ExactMarginalLogLikelihood

    model.train()
    likelihood.train()
    mll = ExactMarginalLogLikelihood(likelihood, model)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
    for _ in range(num_steps):
        optimizer.zero_grad()
        loss = -mll(model(train_x), train_y)
        (loss.sum() if loss.ndim != 0 else loss).backward()
        optimizer.step()
    model.eval()
    likelihood.eval()


def fit_gp(
    x_obs_list: list[float] | list[list[float]],
    y_obs_list: list[float] | list[list[float]],
    num_dim: int,
    *,
    yvar_obs_list: list[float] | None = None,
    num_steps: int = 50,
) -> GPFitResult:
    import numpy as np

    x = np.asarray(x_obs_list, dtype=float)
    y = np.asarray(y_obs_list, dtype=float)
    n, is_multi = x.shape[0], y.ndim == 2 and y.shape[1] > 1
    if n == 0:
        return (
            GPFitResult(
                model=None,
                likelihood=None,
                y_mean=np.zeros(y.shape[1]),
                y_std=np.ones(y.shape[1]),
            )
            if is_multi
            else GPFitResult(model=None, likelihood=None, y_mean=0.0, y_std=1.0)
        )
    if n == 1 and is_multi:
        return GPFitResult(
            model=None,
            likelihood=None,
            y_mean=y[0].copy(),
            y_std=np.ones(int(y.shape[1]), dtype=float),
        )
    gp_data = _prepare_gp_data(x_obs_list, y_obs_list, yvar_obs_list)
    model, likelihood = _build_gp_model(
        gp_data.train_x,
        gp_data.train_y,
        gp_data.is_multi,
        num_dim,
        yvar_obs_list=yvar_obs_list,
        gp_y_std=gp_data.y_std,
        y=gp_data.y_raw,
    )
    _init_gp_hyperparams(
        model,
        gp_data.is_multi,
        num_dim,
        int(gp_data.y_raw.shape[1]) if gp_data.is_multi else None,
        gp_data.train_x.dtype,
    )
    _train_gp(model, likelihood, gp_data.train_x, gp_data.train_y, num_steps)
    return GPFitResult(
        model=model, likelihood=likelihood, y_mean=gp_data.y_mean, y_std=gp_data.y_std
    )



from __future__ import annotations
import conftest
import numpy as np
import pytest
from enn.turbo.optimizer import create_optimizer
from enn.turbo.config import (
    AcqType,
    CandidateRV,
    CandidateGenConfig,
    ENNFitConfig,
    ENNSurrogateConfig,
    MorboTRConfig,
    MultiObjectiveConfig,
    NoTRConfig,
    OptimizerConfig,
    TurboTRConfig,
    turbo_enn_config,
    turbo_one_config,
    turbo_zero_config,
)


def _make_optimizer(*, bounds, config, rng):
    return create_optimizer(bounds=bounds, config=config, rng=rng)


def test_turbo_fallback_called_during_init_with_observations():
    bounds = np.array([[-1.0, 1.0], [-1.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    config = turbo_zero_config(num_init=10)
    opt = _make_optimizer(bounds=bounds, config=config, rng=rng)
    x1 = opt.ask(num_arms=2)
    y1 = conftest.sphere_objective(x1)
    opt.tell(x1, y1)
    x2 = opt.ask(num_arms=2)
    assert x2.shape == (2, 2)
    init = opt.init_progress
    assert init is not None
    init_idx, num_init = init
    assert init_idx <= num_init


def _run_bo(config: OptimizerConfig, num_steps: int = 15) -> float:
    from enn import create_optimizer

    bounds = np.array([[-1.0, 1.0], [-1.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = create_optimizer(bounds=bounds, config=config, rng=rng)
    best = -np.inf
    for _ in range(num_steps):
        x = opt.ask(num_arms=4)
        y = conftest.sphere_objective(x)
        opt.tell(x, y)
        best = max(best, float(np.max(y)))
    return best


def test_turbo_zero_ask_tell_and_shape():
    from enn import create_optimizer

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = create_optimizer(bounds=bounds, config=turbo_zero_config(), rng=rng)
    x0 = opt.ask(num_arms=4)
    assert x0.shape == (4, 2) and np.all(x0 >= 0.0) and np.all(x0 <= 1.0)
    opt.tell(x0, conftest.sphere_objective(x0))
    x1 = opt.ask(num_arms=4)
    assert x1.shape == (4, 2)


def test_optimizer_accepts_list_bounds():
    from enn import create_optimizer

    opt = create_optimizer(
        bounds=[[0.0, 1.0], [0.0, 1.0]],
        config=turbo_zero_config(),
        rng=np.random.default_rng(0),
    )
    assert opt.ask(num_arms=2).shape == (2, 2)


def test_optimizer_uniform_candidates_never_calls_sobol():
    from unittest import mock
    from enn import create_optimizer
    from enn.turbo.optimizer_config import turbo_zero_config

    def _sobol_raises(*args, **kwargs):
        raise RuntimeError(
            "Sobol should not be constructed for candidate_rv=CandidateRV.UNIFORM"
        )

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    with mock.patch("scipy.stats.qmc.Sobol", side_effect=_sobol_raises):
        opt = create_optimizer(
            bounds=bounds,
            config=turbo_zero_config(candidate_rv=CandidateRV.UNIFORM),
            rng=rng,
        )
        x0 = opt.ask(num_arms=4)
    assert x0.shape == (4, 2)


def test_optimizer_accepts_base_config_as_turbo_zero():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    opt = _make_optimizer(
        bounds=bounds,
        config=OptimizerConfig(),
        rng=np.random.default_rng(0),
    )
    x = opt.ask(num_arms=2)
    assert x.shape == (2, 2)


def test_turbo_one_improves_on_sphere():
    assert _run_bo(turbo_one_config(), num_steps=12) > -0.5


def test_turbo_one_with_y_var_uses_noisy_gp():
    from enn import create_optimizer
    from enn.turbo.turbo_gp_noisy import TurboGPNoisy

    bounds = np.array([[-1.0, 1.0], [-1.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    opt = create_optimizer(bounds=bounds, config=turbo_one_config(), rng=rng)
    for _ in range(5):
        x = opt.ask(num_arms=4)
        y = conftest.sphere_objective(x)
        opt.tell(x, y, rng.uniform(0.01, 0.1, size=y.shape))
    x = opt.ask(num_arms=4)
    assert x.shape == (4, 2)
    assert isinstance(opt._surrogate._model, TurboGPNoisy)


def test_turbo_zero_reasonable_on_sphere():
    assert _run_bo(turbo_zero_config(), num_steps=12) > -1.5


def test_turbo_enn_uses_enn_and_is_reasonable():
    assert _run_bo(turbo_enn_config(), num_steps=12) > -1.5


def test_turbo_enn_with_k_none_fits_hyperparameters():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    opt = _make_optimizer(bounds=bounds, config=turbo_enn_config(), rng=rng)
    x0 = opt.ask(num_arms=4)
    assert x0.shape == (4, 2) and np.all(x0 >= 0.0) and np.all(x0 <= 1.0)
    opt.tell(x0, -np.sum(x0**2, axis=1))
    x1 = opt.ask(num_arms=4)
    assert x1.shape == (4, 2) and np.all(x1 >= 0.0) and np.all(x1 <= 1.0)


def test_turbo_enn_config_scale_x_flag_runs():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    from enn.turbo.optimizer_config import ENNSurrogateConfig

    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_enn_config(enn=ENNSurrogateConfig(scale_x=True)),
        rng=np.random.default_rng(0),
    )
    x0 = opt.ask(num_arms=3)
    opt.tell(x0, -np.sum(x0**2, axis=1))
    assert opt.ask(num_arms=3).shape == (3, 2)


def test_find_x_center_uses_top_k_for_mu_single_objective():
    from enn.turbo.components import PosteriorResult
    from enn.turbo.optimizer_config import ENNSurrogateConfig, TurboTRConfig

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_enn_config(
            enn=ENNSurrogateConfig(k=3),
            trust_region=TurboTRConfig(noise_aware=True),
        ),
        rng=rng,
    )
    x_obs = rng.uniform(0.0, 1.0, size=(10, 2))
    y_obs = np.arange(10, dtype=float)
    seen: dict[str, tuple[int, int]] = {}

    def _predict(x):
        seen["shape"] = x.shape
        mu = np.arange(x.shape[0], dtype=float).reshape(-1, 1)
        return PosteriorResult(mu=mu, sigma=None)

    opt._surrogate.predict = _predict
    for x, y in zip(x_obs, y_obs):
        opt._x_obs.append(x)
        opt._y_obs.append(np.array([y]))
    opt._update_incumbent()
    center = opt._find_x_center(x_obs, y_obs)
    assert center.shape == (2,)
    assert seen["shape"] == (3, 2)


def test_find_x_center_uses_top_k_union_for_multiobjective():
    from enn.turbo.components import PosteriorResult
    from enn.turbo.optimizer_config import ENNSurrogateConfig, MorboTRConfig

    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(0)
    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_enn_config(
            enn=ENNSurrogateConfig(k=3),
            trust_region=MorboTRConfig(
                noise_aware=True,
                multi_objective=MultiObjectiveConfig(num_metrics=2),
            ),
        ),
        rng=rng,
    )
    x_obs = rng.uniform(0.0, 1.0, size=(5, 2))
    y_obs = np.array(
        [
            [10.0, 0.0],
            [9.0, 1.0],
            [0.0, 10.0],
            [1.0, 9.0],
            [2.0, 2.0],
        ],
        dtype=float,
    )
    seen: dict[str, tuple[int, int]] = {}

    def _predict(x):
        seen["shape"] = x.shape
        mu = np.zeros((x.shape[0], 2), dtype=float)
        return PosteriorResult(mu=mu, sigma=None)

    opt._surrogate.predict = _predict
    for x, y in zip(x_obs, y_obs):
        opt._x_obs.append(x)
        opt._y_obs.append(y)
    opt._update_incumbent()
    center = opt._find_x_center(x_obs, y_obs)
    assert center.shape == (2,)
    assert seen["shape"] == (5, 2)


def test_optimizer_with_trailing_obs():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(42)
    for cfg in [
        turbo_one_config(trailing_obs=5),
        turbo_enn_config(trailing_obs=5),
    ]:
        opt = _make_optimizer(bounds=bounds, config=cfg, rng=rng)
        for _ in range(10):
            x = opt.ask(num_arms=2)
            opt.tell(x, -np.sum(x**2, axis=1))
        assert len(opt._x_obs) == 5 and len(opt._y_obs) == 5
        assert opt.ask(num_arms=2).shape == (2, 2)


def test_trailing_obs_includes_incumbent():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    rng = np.random.default_rng(123)
    for cfg in [
        turbo_one_config(trailing_obs=5),
        turbo_enn_config(trailing_obs=5),
    ]:
        opt = _make_optimizer(bounds=bounds, config=cfg, rng=rng)
        for i in range(15):
            x = opt.ask(num_arms=2)
            y = (
                np.array([10.0, 9.0])
                if i == 0
                else np.array([5.0 - i * 0.1, 4.0 - i * 0.1])
            )
            opt.tell(x, y)
        assert opt.tr_obs_count <= 5
        assert opt.ask(num_arms=2).shape == (2, 2)


def test_optimizer_tell_without_yvar():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    opt = _make_optimizer(
        bounds=bounds, config=turbo_enn_config(), rng=np.random.default_rng(42)
    )
    for _ in range(2):
        x = opt.ask(num_arms=4)
        opt.tell(x, -np.sum(x**2, axis=1))
    x2 = opt.ask(num_arms=4)
    assert x2.shape == (4, 2) and np.all(x2 >= 0.0) and np.all(x2 <= 1.0)


def test_optimizer_yvar_policy_enforced():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    opt = _make_optimizer(
        bounds=bounds, config=turbo_one_config(), rng=np.random.default_rng(0)
    )
    x0 = opt.ask(num_arms=2)
    y0 = -np.sum(x0**2, axis=1)
    opt.tell(x0, y0, 0.1 * np.ones_like(y0))
    x1 = opt.ask(num_arms=2)
    with pytest.raises(ValueError, match="y_var must be provided"):
        opt.tell(x1, -np.sum(x1**2, axis=1))
    opt2 = _make_optimizer(
        bounds=bounds, config=turbo_one_config(), rng=np.random.default_rng(0)
    )
    x0 = opt2.ask(num_arms=2)
    opt2.tell(x0, -np.sum(x0**2, axis=1))
    x1 = opt2.ask(num_arms=2)
    with pytest.raises(ValueError, match="y_var must be omitted"):
        opt2.tell(x1, -np.sum(x1**2, axis=1), 0.1 * np.ones(2))


def test_turbo_one_trust_region_update_is_noise_robust_to_spikes():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_one_config(
            num_init=1,
            num_candidates=16,
            trust_region=TurboTRConfig(noise_aware=True),
        ),
        rng=np.random.default_rng(0),
    )
    opt.ask(num_arms=1)
    opt.tell(np.zeros((1, 2)), np.array([0.0]), y_var=np.array([1e6]))
    opt.ask(num_arms=1)
    opt.tell(np.ones((1, 2)), np.array([100.0]), y_var=np.array([1e6]))
    assert opt._tr_state.best_value < 100.0


def test_turbo_enn_tr_values_do_not_require_full_history_denoising():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    from enn.turbo.config import (
        AcqType,
        CandidateGenConfig,
        ENNSurrogateConfig,
    )

    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_enn_config(
            enn=ENNSurrogateConfig(k=3),
            candidates=CandidateGenConfig(
                num_candidates=lambda *, num_dim, num_arms: 16
            ),
            num_init=1,
            acq_type=AcqType.PARETO,
        ),
        rng=np.random.default_rng(0),
    )
    opt.ask(num_arms=1)
    opt.tell(np.zeros((1, 2)), np.array([0.0]), y_var=np.array([1e6]))
    opt.ask(num_arms=1)
    opt.tell(np.ones((1, 2)), np.array([100.0]), y_var=np.array([1e6]))
    tr_vals = np.asarray(opt._y_tr_list, dtype=float)
    if tr_vals.ndim == 2:
        tr_vals = tr_vals[:, 0]
    assert tr_vals.shape == (2,)
    assert np.allclose(tr_vals, np.array([0.0, 100.0]))


def test_optimizer_no_trust_region_bounds_are_full_box():
    bounds = np.array([[-2.0, 2.0], [-1.0, 1.0], [0.0, 3.0]], dtype=float)
    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_zero_config(
            trust_region=NoTRConfig(), num_init=1, num_candidates=16
        ),
        rng=np.random.default_rng(0),
    )
    x0 = opt.ask(num_arms=1)
    opt.tell(x0, np.array([1.0]))
    x_center = np.array([0.25, 0.5, 0.75], dtype=float)
    lb, ub = opt._tr_state.compute_bounds_1d(x_center)
    assert np.allclose(lb, 0.0) and np.allclose(ub, 1.0)


def test_optimizer_morbo_multi_objective():
    num_dim, num_metrics = 2, 2
    bounds = np.array([[0.0, 1.0]] * num_dim, dtype=float)
    rng = np.random.default_rng(42)
    opt = _make_optimizer(
        bounds=bounds,
        config=turbo_enn_config(
            enn=ENNSurrogateConfig(
                k=2,
                fit=ENNFitConfig(num_fit_samples=1, num_fit_candidates=1),
            ),
            trust_region=MorboTRConfig(
                multi_objective=MultiObjectiveConfig(num_metrics=num_metrics)
            ),
            num_init=1,
            candidates=CandidateGenConfig(
                num_candidates=lambda *, num_dim, num_arms: 2
            ),
            acq_type=AcqType.THOMPSON,
        ),
        rng=rng,
    )
    x = opt.ask(num_arms=1)
    assert x.shape == (1, num_dim)
    y = rng.uniform(0.0, 1.0, size=(1, num_metrics))
    y_est = opt.tell(x, y)
    assert y_est.shape == (1, num_metrics)
    tr_len = opt.tr_length
    assert tr_len is not None and 0.0 < tr_len <= 1.0



from __future__ import annotations
import numpy as np
import pytest
from enn.turbo.turbo_optimizer_utils import (
    reset_timing,
    sobol_seed_for_state,
    validate_tell_inputs,
    trim_trailing_observations,
)
from enn.turbo.types.telemetry import Telemetry
from enn.turbo.turbo_utils import (
    torch_seed_context,
    get_gp_posterior_suppress_warning,
)


def test_sobol_seed_for_state_deterministic():
    result1 = sobol_seed_for_state(12345, restart_generation=0, n_obs=10, num_arms=4)
    result2 = sobol_seed_for_state(12345, restart_generation=0, n_obs=10, num_arms=4)
    assert result1 == result2


def test_sobol_seed_for_state_changes_with_inputs():
    base = sobol_seed_for_state(12345, restart_generation=0, n_obs=10, num_arms=4)
    diff_seed = sobol_seed_for_state(54321, restart_generation=0, n_obs=10, num_arms=4)
    diff_restart = sobol_seed_for_state(
        12345, restart_generation=1, n_obs=10, num_arms=4
    )
    diff_obs = sobol_seed_for_state(12345, restart_generation=0, n_obs=20, num_arms=4)
    diff_arms = sobol_seed_for_state(12345, restart_generation=0, n_obs=10, num_arms=8)
    assert (
        base != diff_seed
        and base != diff_restart
        and base != diff_obs
        and base != diff_arms
    )


def test_validate_tell_inputs_valid_2d():
    x = np.random.randn(10, 3)
    y = np.random.randn(10, 2)
    y_var = np.random.rand(10, 2)
    result = validate_tell_inputs(x, y, y_var, num_dim=3)
    assert result.x.shape == (10, 3) and result.y.shape == (10, 2)
    assert result.y_var.shape == (10, 2) and result.num_metrics == 2


def test_validate_tell_inputs_valid_1d():
    x = np.random.randn(10, 3)
    y = np.random.randn(10)
    result = validate_tell_inputs(x, y, None, num_dim=3)
    assert (
        result.x.shape == (10, 3)
        and result.y.shape == (10,)
        and result.num_metrics == 1
    )


def test_validate_tell_inputs_invalid_x_shape():
    with pytest.raises(ValueError):
        validate_tell_inputs(
            np.random.randn(10, 4), np.random.randn(10), None, num_dim=3
        )


def test_validate_tell_inputs_mismatched_y():
    with pytest.raises(ValueError):
        validate_tell_inputs(
            np.random.randn(10, 3), np.random.randn(5), None, num_dim=3
        )


def test_validate_tell_inputs_invalid_y_shape():
    with pytest.raises(ValueError):
        validate_tell_inputs(
            np.random.randn(10, 3), np.random.randn(10, 2, 3), None, num_dim=3
        )


def test_trim_trailing_observations_no_trim_needed():
    x_list = [[0.1, 0.2]] * 5
    y_list = [1.0] * 5
    y_tr = [1.0] * 5
    yvar = [0.1] * 5
    incumbent = np.array([0])
    result = trim_trailing_observations(
        x_list, y_list, y_tr, yvar, trailing_obs=10, incumbent_indices=incumbent
    )
    assert len(result.x_obs) == 5


def test_trim_trailing_observations_trims():
    x_list = [[i, i] for i in range(20)]
    y_list = list(range(20))
    y_tr = list(range(20))
    yvar = [0.1] * 20
    incumbent = np.array([0])
    result = trim_trailing_observations(
        x_list, y_list, y_tr, yvar, trailing_obs=5, incumbent_indices=incumbent
    )
    assert len(result.x_obs) <= 5 and 0 in [row[0] for row in result.x_obs]


def test_telemetry_dataclass():
    t = Telemetry(dt_fit=0.5, dt_sel=0.3)
    assert t.dt_fit == 0.5 and t.dt_sel == 0.3


@pytest.mark.parametrize("seed1,seed2,should_match", [(42, 42, True), (42, 43, False)])
def test_torch_seed_context(seed1, seed2, should_match):
    import torch

    with torch_seed_context(seed1):
        val1 = torch.randn(3).tolist()
    with torch_seed_context(seed2):
        val2 = torch.randn(3).tolist()
    assert (val1 == val2) == should_match


def test_get_gp_posterior_suppress_warning_basic():
    from enn.turbo.turbo_gp_fit import fit_gp
    import torch

    x = [[0.1, 0.2], [0.3, 0.4], [0.5, 0.6], [0.7, 0.8]]
    y = [1.0, 2.0, 3.0, 4.0]
    gp_result = fit_gp(x, y, num_dim=2, num_steps=10)
    if gp_result.model is not None:
        x_torch = torch.tensor([[0.2, 0.3]], dtype=torch.float64)
        result = get_gp_posterior_suppress_warning(gp_result.model, x_torch)
        assert result is not None


def test_reset_timing():
    class Obj:
        _dt_fit = 1.0
        _dt_gen = 2.0
        _dt_sel = 3.0

    o = Obj()
    reset_timing(o)
    assert o._dt_fit == 0.0 and o._dt_gen == 0.0 and o._dt_sel == 0.0



import numpy as np
from enn.benchmarks.ackley_core import ackley_core


def test_ackley_core_1d():
    x = np.array([1.0, 2.0])
    result = ackley_core(x)
    assert result.shape == (1,)


def test_ackley_core_2d():
    x = np.array([[1.0, 2.0], [0.0, 0.0]])
    result = ackley_core(x)
    assert result.shape == (2,)
    assert result[1] > result[0]



from __future__ import annotations
import numpy as np
import pytest
from enn.turbo.components.acquisition import (
    HnRAcqOptimizer,
    ParetoAcqOptimizer,
    RandomAcqOptimizer,
    ThompsonAcqOptimizer,
    UCBAcqOptimizer,
)
from enn.turbo.components.builder import build_acquisition_optimizer, build_surrogate
from enn.turbo.components.protocols import (
    PosteriorResult,
    SurrogateResult,
)
from enn.turbo.components.surrogates import ENNSurrogate, GPSurrogate, NoSurrogate
from enn.turbo.config import (
    ENNSurrogateConfig,
    NoTRConfig,
    turbo_enn_config,
    turbo_one_config,
    turbo_zero_config,
)
from enn.turbo.optimizer import Optimizer, create_optimizer


def _make_test_data(n: int = 4, d: int = 2):
    x = np.array([[0.2, 0.3], [0.5, 0.5], [0.7, 0.8], [0.1, 0.9]])[:n, :d]
    y = np.array([0.5, 0.7, 0.3, 0.6])[:n]
    return x, y


def _make_candidates(n: int = 4, d: int = 2):
    return np.array([[0.1, 0.2], [0.3, 0.4], [0.5, 0.6], [0.7, 0.8]])[:n, :d]


def _fit_gp_surrogate(rng):
    surrogate = GPSurrogate()
    x, y = _make_test_data()
    surrogate.fit(x, y, None, num_steps=10, rng=rng)
    return surrogate


def _fit_no_surrogate(rng, n: int = 2):
    surrogate = NoSurrogate()
    x, y = _make_test_data(n=n)
    surrogate.fit(x, y, None, num_steps=0, rng=rng)
    return surrogate


def test_surrogate_result():
    result = SurrogateResult(model="test_model", lengthscales=np.array([1.0, 2.0]))
    assert result.model == "test_model"
    assert np.allclose(result.lengthscales, [1.0, 2.0])


def test_surrogate_result_defaults():
    result = SurrogateResult(model=None)
    assert result.model is None
    assert result.lengthscales is None


def test_posterior_result():
    mu = np.array([[1.0], [2.0]])
    sigma = np.array([[0.1], [0.2]])
    result = PosteriorResult(mu=mu, sigma=sigma)
    assert np.allclose(result.mu, mu)
    assert np.allclose(result.sigma, sigma)


def test_posterior_result_no_sigma():
    mu = np.array([[1.0], [2.0]])
    result = PosteriorResult(mu=mu)
    assert np.allclose(result.mu, mu)
    assert result.sigma is None


def test_no_surrogate_fit():
    surrogate = NoSurrogate()
    x, y = _make_test_data(n=2)
    result = surrogate.fit(x, y, None, num_steps=0, rng=np.random.default_rng(0))
    assert result.model is None
    assert result.lengthscales is None


def test_no_surrogate_predict():
    rng = np.random.default_rng(0)
    surrogate = _fit_no_surrogate(rng)
    x, _ = _make_test_data(n=2)
    posterior = surrogate.predict(x)
    assert posterior.mu.shape == (2, 1)
    assert posterior.sigma is None


def test_no_surrogate_sample():
    rng = np.random.default_rng(42)
    surrogate = _fit_no_surrogate(rng)
    x, _ = _make_test_data(n=2)
    samples = surrogate.sample(x, 3, rng)
    assert samples.shape == (3, 2, 1)


def test_gp_surrogate_init():
    surrogate = GPSurrogate()
    assert surrogate._model is None


def test_gp_surrogate_fit():
    surrogate = GPSurrogate()
    x, y = _make_test_data()
    rng = np.random.default_rng(42)
    result = surrogate.fit(x, y, None, num_steps=10, rng=rng)
    assert result.model is not None
    assert result.lengthscales is not None
    assert result.lengthscales.shape == (2,)


def test_gp_surrogate_predict():
    rng = np.random.default_rng(42)
    surrogate = _fit_gp_surrogate(rng)
    x, _ = _make_test_data()
    posterior = surrogate.predict(x)
    assert posterior.mu.shape == (4, 1)
    assert posterior.sigma.shape == (4, 1)


def test_gp_surrogate_predict_not_fitted_raises():
    surrogate = GPSurrogate()
    x = np.array([[0.5, 0.5]])
    with pytest.raises(RuntimeError, match="fitted"):
        surrogate.predict(x)


def test_enn_surrogate_init():
    config = ENNSurrogateConfig(k=5)
    surrogate = ENNSurrogate(config)
    assert surrogate._config.k == 5


def test_enn_surrogate_fit():
    config = ENNSurrogateConfig(k=3)
    surrogate = ENNSurrogate(config)
    x, y = _make_test_data()
    rng = np.random.default_rng(42)
    result = surrogate.fit(x, y, None, num_steps=0, rng=rng)
    assert result.model is not None


def test_enn_surrogate_predict():
    config = ENNSurrogateConfig(k=3)
    surrogate = ENNSurrogate(config)
    x, y = _make_test_data()
    rng = np.random.default_rng(42)
    surrogate.fit(x, y, None, num_steps=0, rng=rng)
    posterior = surrogate.predict(x)
    assert posterior.mu.shape == (4, 1)


def test_no_surrogate_get_incumbent_candidate_indices():
    surrogate = NoSurrogate()
    y_obs = np.array([0.1, 0.3, -0.2], dtype=float)
    indices = surrogate.get_incumbent_candidate_indices(y_obs)
    assert np.array_equal(indices, np.array([0, 1, 2], dtype=int))


def test_gp_surrogate_get_incumbent_candidate_indices():
    surrogate = GPSurrogate()
    y_obs = np.array([0.2, -0.1, 0.4], dtype=float)
    indices = surrogate.get_incumbent_candidate_indices(y_obs)
    assert np.array_equal(indices, np.array([0, 1, 2], dtype=int))


def test_enn_surrogate_get_incumbent_candidate_indices_top_k():
    config = ENNSurrogateConfig(k=2)
    surrogate = ENNSurrogate(config)
    y_obs = np.array([0.1, 0.9, 0.3, 0.8], dtype=float)
    indices = surrogate.get_incumbent_candidate_indices(y_obs)
    assert np.array_equal(np.sort(indices), np.array([1, 3], dtype=int))


@pytest.mark.parametrize(
    "optimizer_cls,surrogate_fn",
    [
        (RandomAcqOptimizer, _fit_no_surrogate),
        (ThompsonAcqOptimizer, _fit_no_surrogate),
    ],
)
def test_acquisition_optimizer_select_with_no_surrogate(optimizer_cls, surrogate_fn):
    optimizer = optimizer_cls()
    x_cand = _make_candidates()
    rng = np.random.default_rng(42)
    surrogate = surrogate_fn(rng)
    selected = optimizer.select(x_cand, 2, surrogate, rng)
    assert selected.shape == (2, 2)


def test_ucb_acq_optimizer_init():
    optimizer = UCBAcqOptimizer(beta=2.0)
    assert optimizer._beta == 2.0


def test_ucb_acq_optimizer_select():
    optimizer = UCBAcqOptimizer(beta=1.0)
    rng = np.random.default_rng(42)
    surrogate = _fit_gp_surrogate(rng)
    x_cand = _make_candidates()
    selected = optimizer.select(x_cand, 2, surrogate, rng)
    assert selected.shape == (2, 2)


def test_pareto_acq_optimizer_select():
    optimizer = ParetoAcqOptimizer()
    rng = np.random.default_rng(42)
    config = ENNSurrogateConfig(k=3)
    surrogate = ENNSurrogate(config)
    x, y = _make_test_data()
    surrogate.fit(x, y, None, num_steps=0, rng=rng)
    x_cand = _make_candidates()
    selected = optimizer.select(x_cand, 2, surrogate, rng)
    assert selected.shape == (2, 2)


def test_hnr_acq_optimizer_init():
    base = UCBAcqOptimizer(beta=1.0)
    optimizer = HnRAcqOptimizer(base, num_iterations=10)
    assert optimizer._num_iterations == 10


@pytest.mark.parametrize(
    "base_cls,surrogate_fn",
    [
        (UCBAcqOptimizer, _fit_gp_surrogate),
        (ThompsonAcqOptimizer, _fit_no_surrogate),
    ],
)
def test_hnr_acq_optimizer_select(base_cls, surrogate_fn):
    if base_cls == UCBAcqOptimizer:
        base = base_cls(beta=1.0)
    else:
        base = base_cls()
    optimizer = HnRAcqOptimizer(base, num_iterations=5)
    rng = np.random.default_rng(42)
    surrogate = surrogate_fn(rng)
    x_cand = _make_candidates()
    selected = optimizer.select(x_cand, 2, surrogate, rng)
    assert selected.shape == (2, 2)


@pytest.mark.parametrize(
    "config_fn,expected_cls",
    [
        (turbo_one_config, GPSurrogate),
        (turbo_zero_config, NoSurrogate),
        (turbo_enn_config, ENNSurrogate),
    ],
)
def test_build_surrogate(config_fn, expected_cls):
    config = config_fn()
    surrogate = build_surrogate(config)
    assert isinstance(surrogate, expected_cls)


@pytest.mark.parametrize(
    "config_fn,expected_cls",
    [
        (turbo_zero_config, RandomAcqOptimizer),
        (turbo_one_config, ThompsonAcqOptimizer),
        (turbo_enn_config, ParetoAcqOptimizer),
    ],
)
def test_build_acquisition_optimizer(config_fn, expected_cls):
    config = config_fn()
    optimizer = build_acquisition_optimizer(config)
    assert isinstance(optimizer, expected_cls)


def test_surrogate_protocol_compliance():
    for cls in [NoSurrogate, GPSurrogate]:
        assert hasattr(cls, "fit")
        assert hasattr(cls, "predict")
        assert hasattr(cls, "sample")


def test_acquisition_optimizer_protocol_compliance():
    for cls in [
        RandomAcqOptimizer,
        ThompsonAcqOptimizer,
        UCBAcqOptimizer,
        ParetoAcqOptimizer,
    ]:
        assert hasattr(cls, "select")


def test_trust_region_protocol_from_config_build():
    tr_config = NoTRConfig()
    rng = np.random.default_rng(42)
    tr = tr_config.build(num_dim=3, rng=rng)
    assert hasattr(tr, "length")
    assert hasattr(tr, "compute_bounds_1d")
    assert hasattr(tr, "update")
    assert hasattr(tr, "needs_restart")
    assert hasattr(tr, "restart")


def test_optimizer_init():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    config = turbo_zero_config(num_init=4)
    rng = np.random.default_rng(42)
    surrogate = NoSurrogate()
    acq_optimizer = RandomAcqOptimizer()
    opt = Optimizer(
        bounds=bounds,
        config=config,
        rng=rng,
        surrogate=surrogate,
        acquisition_optimizer=acq_optimizer,
    )
    assert opt is not None
    assert opt._num_dim == 2
    init = opt.init_progress
    assert init is not None
    assert init == (0, 4)


def test_optimizer_init_via_factory():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    config = turbo_zero_config(num_init=4)
    rng = np.random.default_rng(42)
    opt = create_optimizer(bounds=bounds, config=config, rng=rng)
    assert opt is not None
    assert isinstance(opt, Optimizer)


def test_optimizer_fallback_during_init():
    bounds = np.array([[0.0, 1.0], [0.0, 1.0]], dtype=float)
    config = turbo_one_config(num_init=10, num_candidates=16)
    rng = np.random.default_rng(42)
    opt = create_optimizer(bounds=bounds, config=config, rng=rng)
    x1 = opt.ask(num_arms=2)
    y1 = -np.sum(x1**2, axis=1)
    opt.tell(x1, y1)
    x2 = opt.ask(num_arms=2)
    assert x2.shape == (2, 2)
    init = opt.init_progress
    assert init is not None
    init_idx, num_init = init
    assert init_idx < num_init
